# This file is part of pymadcad,  distributed under license LGPL v3

'''	Display module of pymadcad
	
	This module provides a render pipeline system featuring:

	- Class `Scene` to gather the data to render
	- Widget `View` that actually renders the scene 
	- The display protocol, that allows any object to define its `Display` subclass to be rendered in a scene.

	The view is for window integration and user interaction. `Scene` is only to manage the objects to render . Almost all madcad data types can be rendered to scenes being converted into an appropriate subclass of `Display`. Since the conversion from madcad data types into display instance is automatically handled via the *display protocol*, you usually don't need to deal with displays directly.

	
	Display protocol
	----------------
		A displayable is an object that implements the signature of Display:
		
			class display:
				box (Box)                      # delimiting the display, can be an empty or invalid box
				world (fmat4)                  # local transformation
				
				__getitem__                    # access to subdisplays if there is
				stack(scene)                   # rendering routines (can be methods, or any callable)
				
				duplicate(src,dst)             # copy the display object for an other scene if possible
				update(scene,displayable)     # upgrade the current display to represent the given displayable
				control(...)                   # handle events
		
		For more details, see class Display below
		
	WARNING
	-------
		As the GPU native precision is f4 (float 32 bits), all the vector stuff regarding rendering is made using simple precision types: `fvec3, fvec4, fmat3, fmat4, ...`
	
	NOTE
	----
		There is some restrictions using the widget. This is due to some Qt limitations (and design choices), that Qt is using separated opengl contexts for each independent widgets or window.
		
		- a View should not be reparented once displayed
		- a View can't share a scene with Views from an other window
		- to share a Scene between Views, you must activate 
				QCoreApplication.setAttribute(Qt.AA_ShareOpenGLContexts)
'''

import traceback
from copy import copy, deepcopy
from operator import itemgetter

import moderngl as mgl
import numpy.core as np
from PIL import Image
from PyQt5.QtCore import QEvent, QPoint, Qt
from PyQt5.QtGui import (QFocusEvent, QInputEvent, QKeyEvent, QMouseEvent,
                         QSurfaceFormat, QTouchEvent)
from PyQt5.QtWidgets import QApplication, QOpenGLWidget, QWidget

from . import settings
from .common import resourcedir
from .mathutils import *
from .nprint import nprint

# minimum opengl version required by the rendering pipeline
opengl_version = (4,3)
# shared open gl context, None if not yet initialized
global_context = None

def qt_surface_format():
    fmt = QSurfaceFormat()
    fmt.setVersion(*opengl_version)
    fmt.setProfile(QSurfaceFormat.OpenGLContextProfile.CoreProfile)
    fmt.setSamples(4)
    return fmt

def show(scene:dict, interest:Box=None, size=uvec2(400,400), projection=None, navigation=None, **options):
	''' 
		Easy and convenient way to create a window containing a `View` on a created `Scene`
		
		If a Qt app is not already running, the functions returns when the window has been closed and all GUI destroyed
		
		Parameters:
			scene:     a mapping (dict or list) giving the objects to render in the scene
			interest:  the region of interest to zoom on at the window initialization
			size:      the window size (pixel)
			options:   options to set in `Scene.options`
		
		Tip:
			For integration in a Qt window or to manipulate the view, you should directly use `View`
	'''
	global global_context

	if isinstance(scene, list):	scene = dict(enumerate(scene))
	# retro-compatibility fix, shall be removed in future versions
	if 'options' in options:	options.update(options['options'])
	if not isinstance(scene, Scene):	scene = Scene(scene, options)
	
	QSurfaceFormat.setDefaultFormat(qt_surface_format())

	app = QApplication.instance()
	created = False
	if not app:
		import sys
		QApplication.setAttribute(Qt.AA_ShareOpenGLContexts, True)
		app = QApplication(sys.argv)
		global_context = None
		created = True

	# use the Qt color scheme if specified
	if settings.display['system_theme']: 
		settings.use_qt_colors()

	# create the scene as a window
	view = View(scene, projection=projection, navigation=navigation)
	view.resize(*size)
	view.show()

	# make the camera see everything
	if not interest:	
		interest = view.scene.box()
	view.center(interest.center)
	view.adjust(interest)

	if created:
		err = app.exec()
		if err != 0:	print('error: Qt exited with code', err)

def render(scene, options=None, interest:Box=None, navigation=None, projection=None, size=uvec2(400,400)):
	''' Shortcut to render the given objects to an image, returns a PIL Image
	
		For repeated renderings or view manipualtion, you should directly use `Offscreen`
		
		NOTE:
			the system theme colors cannot be automatically loaded since no running QApplication is assumed in the function
	'''
	if isinstance(scene, list):	scene = dict(enumerate(scene))
	if not isinstance(scene, Scene):	scene = Scene(scene, options)

	# create the scene and an offscreen renderer
	view = Offscreen(scene, size, navigation=navigation, projection=projection)

	# load objects in the scene, so the scene's box can be computed
	with scene.ctx:
		scene.dequeue()

	# make the camera see everything
	if not navigation:
		if not interest:
			interest = view.scene.box()
		view.center(interest.center)
		view.adjust(interest)

	return view.render()





class Display:
	''' Blanket implementation for displays.
		This class signature is exactly the display protocol specification
		
		Attributes:
		
			world(fmat4):  matrix from local space to parent space
			box(Box):      boudingbox of the display in local space
			
		These attributes are variable members by default but can be overriden as properties if needed.
	'''
	
	# mendatory part of the protocol
	
	box = Box(center=0, width=fvec3(-inf))	# to inform the scene and the view of the object size
	world = fmat4(1)		# set by the display containing this one if it is belonging to a group
	
	def display(self, scene) -> 'self':
		''' Displays are obviously displayable as themselves '''
		return self
	def stack(self, scene) -> '[(key, target, priority, callable)]':
		''' Rendering functions to insert in the renderpipeline.
		
			The expected result can be any iterable providing tuples `(key, target, priority, callable)` such as:
			
			:key:	    a tuple with the successive keys in the displays tree, most of the time implementers set it to `()`  because it doesn't belong to a subpart of the Display.
			:target:    the name of the render target in the view that will be rendered (see View)
			:priority:  a float that is used to insert the callable at the proper place in the rendering stack
			:callable:  a function that renders, signature is  `func(view)`			
			
			The view contains the uniforms, rendering targets and the scene for common resources
		'''
		return ()
	def duplicate(self, src, dst) -> 'display/None':
		''' Duplicate the display for an other scene (other context) but keeping the same memory buffers when possible.
			
			Return None if not possible or not implemented.
		'''
		return None
	def __getitem__(self, key) -> 'display':
		''' Get a subdisplay by its index/key in this display (like in a scene) '''
		raise IndexError('{} has no sub displays'.format(type(self).__name__))
	def update(self, scene, displayable) -> bool:
		''' Update the current displays internal datas with the given displayable .
			
			If the display cannot be upgraded, it must return False to be replaced by a fresh new display created from the displayable
		'''
		return False
	
	# optional part for usage with Qt
	
	selected = False
	
	def control(self, view, key, sub, evt: 'QEvent'):
		''' Handle input events occuring on the area of this display (or of one of its subdisplay).
			For subdisplay events, the parents control functions are called first, and the sub display controls are called only if the event is not accepted by parents
			
			Parameters:
				key:    the key path for the current display
				sub:    the key path for the subdisplay
				evt:    the Qt event (see Qt doc)
		'''
		pass

def qt_2_glm(v):
	if isinstance(v, (QPoint, QPointF)):	return vec2(v.x(), v.y())
	elif isinstance(v, (QSize, QSizeF)):	return vec2(v.width(), v.height())
	else:
		raise TypeError("can't convert {} to vec2".format(type(v).__name__))

def navigation_tool(dispatcher, view):
	''' Internal navigation tool '''	
	ctrl = alt = slow = False
	nav = curr = None
	moving = False
	hastouched = False
	while True:
		evt = yield
		
		if isinstance(evt, QKeyEvent):
			k = evt.key()
			press = evt.type() == QEvent.KeyPress
			if	 k == Qt.Key_Control:	ctrl = press
			elif k == Qt.Key_Alt:		alt = press
			elif k == Qt.Key_Shift:		slow = press
		elif evt.type() == QEvent.MouseButtonPress:
			modifiers = QApplication.queryKeyboardModifiers()
			ctrl = bool(modifiers & Qt.ControlModifier)
			alt = bool(modifiers & Qt.AltModifier)
			shift = bool(modifiers & Qt.ShiftModifier)
		if isinstance(evt, QKeyEvent) or evt.type() == QEvent.MouseButtonPress:
			if ctrl and alt:		curr = 'zoom'
			elif ctrl:				curr = 'pan'
			elif alt:				curr = 'rotate'
			else:					curr = None
		
		if isinstance(evt, QKeyEvent):
			evt.ignore()	# ignore the keys to pass shortcuts to parents
		elif evt.type() == QEvent.MouseButtonPress:
			last = evt.pos()
			if evt.button() == Qt.MiddleButton:
				nav = 'rotate'
			else:
				nav = curr
			# prevent any scene interaction
			if nav:
				evt.accept()
		elif evt.type() == QEvent.MouseMove:
			if nav:
				moving = True
				gap = evt.pos() - last
				dx = gap.x()/view.height()
				dy = gap.y()/view.height()
				if nav == 'pan':		view.navigation.pan(dx, dy)
				elif nav == 'rotate':	view.navigation.rotate(dx, dy, 0)
				elif nav == 'zoom':		
					middle = QPoint(view.width(), view.height())/2
					f = (	(last-middle).manhattanLength()
						/	(evt.pos()-middle).manhattanLength()	)
					view.navigation.zoom(f)
				last = evt.pos()
				view.update()
				evt.accept()
		elif evt.type() == QEvent.MouseButtonRelease:
			if moving:
				moving = False
				evt.accept()
		elif evt.type() == QEvent.Wheel:
			view.navigation.zoom(exp(-evt.angleDelta().y()/(8*90)))	# the 8 factor is there because of the Qt documentation
			view.update()
			evt.accept()
	
				
		elif isinstance(evt, QTouchEvent):
			nav = None
			pts = evt.touchPoints()
			# view rotation
			if len(pts) == 2:
				startlength = (pts[0].lastPos()-pts[1].lastPos()).manhattanLength()
				zoom = startlength / (pts[0].pos()-pts[1].pos()).manhattanLength()
				displt = (	(pts[0].pos()+pts[1].pos()) /2 
						-	(pts[0].lastPos()+pts[1].lastPos()) /2 ) /view.height()
				dc = pts[0].pos() - pts[1].pos()
				dl = pts[0].lastPos() - pts[1].lastPos()
				rot = atan2(dc.y(), dc.x()) - atan2(dl.y(), dl.x())
				view.navigation.zoom(zoom)
				view.navigation.rotate(displt.x(), displt.y(), rot)
				hastouched = True
				view.update()
				evt.accept()
			# view translation
			elif len(pts) == 3:
				lc = (	pts[0].lastPos() 
					+	pts[1].lastPos() 
					+	pts[2].lastPos() 
					)/3
				lr = (	(pts[0].lastPos() - lc) .manhattanLength()
					+	(pts[1].lastPos() - lc) .manhattanLength()
					+	(pts[2].lastPos() - lc) .manhattanLength()
					)/3
				cc = (	pts[0].pos() 
					+	pts[1].pos() 
					+	pts[2].pos() 
					)/3
				cr = (	(pts[0].pos() - cc) .manhattanLength()
					+	(pts[1].pos() - cc) .manhattanLength()
					+	(pts[2].pos() - cc) .manhattanLength()
					)/3
				zoom = lr / cr
				displt = (cc - lc)  /view.height()
				view.navigation.zoom(zoom)
				view.navigation.pan(displt.x(), displt.y())
				hastouched = True
				view.update()
				evt.accept()
			# finish a gesture
			elif evt.type() in (QEvent.TouchEnd, QEvent.TouchUpdate):
				evt.accept()
				

class Turntable:
	''' Navigation rotating on yaw and pitch around a center 
	
		Object used as `View.navigation`
	'''
	def __init__(self, center:fvec3=0, distance:float=1, yaw:float=0, pitch:float=0):
		self.center = fvec3(center)
		self.yaw = yaw
		self.pitch = pitch
		self.distance = distance
		self.tool = navigation_tool
		
	def rotate(self, dx, dy, dz):
		if abs(self.pitch) > 3.2:	dx = -dx
		self.yaw += dx*pi
		self.pitch += dy*pi
		if self.pitch > pi:	self.pitch -= 2*pi
		if self.pitch < -pi: self.pitch += 2*pi
	def pan(self, dx, dy):
		mat = transpose(fmat3(inverse(fquat(fvec3(pi/2-self.pitch, 0, -self.yaw)))))
		self.center += ( mat[0] * -dx + mat[1] * dy) * self.distance/2
	def zoom(self, f):
		self.distance *= f
	
	def matrix(self) -> fmat4:
		# build rotation from view euler angles
		rot = inverse(fquat(fvec3(pi/2-self.pitch, 0, -self.yaw)))
		mat = translate(fmat4(rot), -self.center)
		mat[3][2] -= self.distance
		return mat

class Orbit:
	''' Navigation rotating on the 3 axis around a center.
	
		Object used as `View.navigation`
	'''
	def __init__(self, center:fvec3=0, distance:float=1, orient:fvec3=fvec3(1,0,0)):
		self.center = fvec3(center)
		self.distance = float(distance)
		self.orient = fquat(orient)
		self.tool = navigation_tool
		
	def rotate(self, dx, dy, dz):
		# rotate from view euler angles
		self.orient = inverse(fquat(fvec3(-dy, -dx, dz) * pi)) * self.orient
	def pan(self, dx, dy):
		x,y,z = transpose(fmat3(self.orient))
		self.center += (fvec3(x) * -dx + fvec3(y) * dy) * self.distance/2
	def zoom(self, f):
		self.distance *= f
	
	def matrix(self) -> fmat4:
		mat = translate(fmat4(self.orient), -self.center)
		mat[3][2] -= self.distance
		return mat


class Perspective:
	''' Object used as `View.projection` 
	
		Attributes:
			fov (float):	field of view (rad), defaulting to `settings.display['field_of_view']`
	'''
	def __init__(self, fov=None):
		self.fov = fov or settings.display['field_of_view']
	def matrix(self, ratio, distance) -> fmat4:
		return perspective(self.fov, ratio, distance*1e-2, distance*1e4)

class Orthographic:
	''' Object used as `View.projection` 
	
		Attributes:
			size (float):  
			
				factor between the distance from camera to navigation center and the zone size to display
				defaulting to `tan(settings.display['field_of_view']/2)`
	'''
	def __init__(self, size=None):
		self.size = size or tan(settings.display['field_of_view']/2)

	def matrix(self, ratio, distance) -> fmat4:
		return fmat4(1/(ratio*distance*self.size), 0, 0, 0,
		            0,       1/(distance*self.size), 0, 0,
		            0,       0,          -2/(distance*(1e3-1e-2)), 0,
		            0,       0,          -(1e3+1e-2)/(1e3-1e-2), 1)


class Scene:
	''' Rendering pipeline for madcad displayable objects 
		
		This class is gui-agnostic, it only relies on OpenGL, and the context has to be created by the user.
		
		When an object is added to the scene, a Display is not immediately created for it, the object is put into the queue and the Display is created at the next render.
		If the object is removed from the scene before the next render, it is dequeued.
		
		Attributes:
		
			ctx:				  moderngl Context (must be the same for all views using this scene)
			resources (dict):    dictionary of scene resources (like textures, shaders, etc) index by name
			options (dict):       dictionary of options for rendering, initialized with a copy of `settings.scene`
			
			displays (dict):      dictionary of items in the scheme `{'name': Display}`
			stacks (list):        lists of callables to render each target `{'target': [(key, priority, callable(view))]}`
			setup (dict):         setup of each rendering target `{'target': callable}`
			
			touched (bool):       flag set to True if the stack must be recomputed at the next render time (there is a change in a Display or in one of its children)
	'''
	
	def __init__(self, objs=(), options=None, ctx=None, setup=None):
		# context variables
		self.ctx = ctx
		self.resources = {}	# context-related resources, shared across displays, but not across contexts (shaders, vertexarrays, ...)
		
		# rendering options
		self.options = deepcopy(settings.scene)
		if options:	self.options.update(options)
		
		# render elements
		self.queue = {}	# list of objects to display, not yet loaded on the GPU
		self.displays = {} # displays created from the inserted objects, associated to their insertion key
		self.stacks = {}	# dict of list of callables, that constitute the render pipeline:  (key,  priority, callable)
		self.setup = setup or {}	# callable for each target
		
		self.touched = False
		self.update(objs)
	
	# methods to manage the rendering pipeline
	
	def add(self, displayable, key=None) -> 'key':
		''' Add a displayable object to the scene, if key is not specified, an unused integer key is used 
			The object is not added to the render pipeline yet, but queued for next rendering.
		'''
		if key is None:
			for i in range(len(self.displays)+len(self.queue)+1):
				if i not in self.displays and i not in self.queue:	key = i
		self.queue[key] = displayable
		return key

	def __setitem__(self, key, value):
		''' Equivalent with self.add with a key '''
		self.queue[key] = value
	def __getitem__(self, key) -> 'display':
		''' Get the displayable for the given key, raise when there is no object or when the object is still in queue. '''
		return self.displays[key]
	def __delitem__(self, key):
		''' Remove an item from the scene, at the root level '''
		if key in self.displays:
			del self.displays[key]
		if key in self.queue:
			del self.queue[key]
		for stack in self.stacks.values():
			for i in reversed(range(len(stack))):
				if stack[i][0][0] == key:
					stack.pop(i)
					
	def item(self, key):
		''' Get the Display associated with the given key, descending the parenting tree 
		
			The parents must all make their children accessible via `__getitem__`
		'''
		disp = self.displays
		for i in range(1,len(key)):
			disp = disp[key[i-1]]
		return disp
	
	def update(self, objs:dict):
		''' Rebuild the scene from a dictionary of displayables 
			Update former displays if possible instead of replacing it
		'''
		self.queue.update(objs)
		self.touch()
	
	def sync(self, objs:dict):
		''' Update the scene from a dictionary of displayables, the former values that cannot be updated are discarded '''
		for key in list(self.displays):
			if key not in objs:
				del self.displays[key]
		self.update(objs)
	
	def touch(self):
		''' Shorthand for `self.touched = True` '''
		self.touched = True
		
	def dequeue(self):
		''' Load all pending objects to insert into the scene.
			This is called automatically by the next `render()` if `touch()` has been called
		'''
		if self.queue:
			with self.ctx:
				self.ctx.finish()
				# update displays
				for key,displayable in self.queue.items():
					try:	
						self.displays[key] = self.display(displayable, self.displays.get(key))
					except:
						print('\ntried to display', object.__repr__(displayable))
						traceback.print_exc()
				self.touched = True
				self.queue.clear()
		
		if self.touched:
			self.restack()
			
	def restack(self):
		''' Update the rendering calls stack from the current scene's displays.
			This is called automatically on `dequeue()`
		'''
		# recreate stacks
		for stack in self.stacks.values():
			stack.clear()
		for key,display in self.displays.items():
			for frame in display.stack(self):
				if len(frame) != 4:
					raise ValueError('wrong frame format in the stack from {}\n\t got {}'.format(display, frame))
				sub,target,priority,func = frame
				if target not in self.stacks:	self.stacks[target] = []
				stack = self.stacks[target]
				stack.append(((key,*sub), priority, func))
		# sort the stack using the specified priorities
		for stack in self.stacks.values():
			stack.sort(key=itemgetter(1))
		self.touched = False
	
	def render(self, view):
		''' Render to the view targets. 
			
			This must be called by the view widget, once the OpenGL context is set.
		'''
		empty = ()
		with self.ctx:
			# apply changes that need opengl runtime
			self.dequeue()
			# render everything
			for target, frame, setup in view.targets:
				view.target = frame
				frame.use()
				frame.clear()
				setup()
				for key, priority, func in self.stacks.get(target,empty):
					func(view)
	
	def box(self):
		''' Computes the boundingbox of the scene, with the current object poses '''
		box = Box(center=fvec3(0), width=fvec3(-inf))
		for display in self.displays.values():
			box.union_update(display.box.transform(display.world))
		return box
	
	def resource(self, name, func=None):
		''' Get a resource loaded or load it using the function func.
			If func is not provided, an error is raised
		'''
		if name in self.resources:	
			return self.resources[name]
		elif callable(func):
			with self.ctx as ctx:  # set the scene context as current opengl context
				res = func(self)
				self.resources[name] = res
				return res
		else:
			raise KeyError("resource {} doesn't exist or is not loaded".format(repr(name)))
					
	def display(self, obj, former=None):
		''' Create a display for the given object for the current scene.
		
			This is the actual function converting objects into displays.
			You don't need to call this method if you just want to add an object to the scene, use add() instead
		'''
		if former and former.update(self, obj):
			return former
		if type(obj) in overrides:
			disp = overrides[type(obj)](self, obj)
		elif hasattr(obj, 'display'):
			if isinstance(obj.display, type):
				disp = obj.display(self, obj)
			elif callable(obj.display):
				disp = obj.display(self)
			else:
				raise TypeError("member 'display' must be a method or a type, on {}".format(type(obj).__name__))
		else:
			raise TypeError('type {} is not displayable'.format(type(obj).__name__))
		
		if not isinstance(disp, Display):
			raise TypeError('the display for {} is not a subclass of Display: {}'.format(type(obj).__name__, type(disp)))
		return disp
	

def displayable(obj):
	''' Return True if the given object has the matching signature to be added to a Scene '''
	return type(obj) in overrides or hasattr(obj, 'display') and callable(obj.display) and not isinstance(obj, type)


class Step(Display):
	''' Simple display holding a rendering stack step 
	
		`Step(target, priority, callable)`
	'''
	__slots__ = 'step',
	def __init__(self, target, priority, callable):	self.step = ((), target, priority, callable)
	def __repr__(self):		return '{}({}, {}, {})'.format(type(self).__name__, self.step[1], self.step[2], self.step[3])
	def stack(self, scene):		return self.step,

class Displayable:
	''' Simple displayable initializing the given Display class with arguments 
		
		At the display creation time, it will simply execute `build(*args, **kwargs)`
	'''
	__slots__ = 'build', 'args', 'kwargs'
	def __init__(self, build, *args, **kwargs):
		self.args, self.kwargs = args, kwargs
		self.build = build
	def __repr__(self):
		return '{}({}, {}, {})'.format(type(self).__name__, repr(self.args[1:-1]), repr(self.kwargs)[1:-1])
	def display(self, scene):
		return self.build(scene, *self.args, **self.kwargs)


def writeproperty(func):
	''' Decorator to create a property that has only an action on variable write '''
	fieldname = '_'+func.__name__
	def getter(self):	return getattr(self, fieldname)
	def setter(self, value):
		setattr(self, fieldname, value)
		func(self, value)
	return property(getter, setter, doc=func.__doc__)

class Group(Display):
	''' A group is like a subscene '''
	def __init__(self, scene, objs:'dict/list'=None, local=1):
		self._world = fmat4(1)
		self._local = fmat4(local)
		self.displays = {}
		if objs:	self.dequeue(scene, objs)
	
	def __getitem__(self, key):
		return self.displays[key]
	def __iter__(self):
		return iter(self.displays.values())
		
	def update(self, scene, objs):
		if isinstance(objs, dict):		objs = objs
		elif hasattr(objs, 'keys'):		objs = dict(objs)
		elif hasattr(objs, '__iter__'):	objs = dict(enumerate(objs))
		else:
			return False
		
		# update displays
		sub = self._world * self._local
		with scene.ctx:
			scene.ctx.finish()
			for key, obj in objs.items():
				if not displayable(obj):	continue
				try:
					self.displays[key] = disp = scene.display(obj, self.displays.get(key))
					disp.world = sub
				except:
					print('tried to display', object.__repr__(obj))
					traceback.print_exc()
			for key in self.displays.keys() - objs.keys():
				del self.displays[key]
		scene.touch()
		return True
	dequeue = update
	
	def stack(self, scene):
		for key,display in self.displays.items():
			for sub,target,priority,func in display.stack(scene):
				yield ((key, *sub), target, priority, func)
	
	@writeproperty
	def local(self, pose):
		''' Pose of the group relatively to its parents '''
		sub = self._world * self._local
		for display in self.displays.values():
			display.world = sub
			
	@writeproperty
	def world(self, world):
		''' Update children's world matrix applying the current pose in addition to world '''
		sub = self._world * self._local
		for display in self.displays.values():
			display.world = sub
			
	@property
	def box(self):
		''' Computes the boundingbox of the scene, with the current object poses '''
		box = Box(center=fvec3(0), width=fvec3(-inf))
		for display in self.displays.values():
			box.union_update(display.box)
		return box.transform(self._local)


# dictionary to store procedures to override default object displays
overrides = {
	list: Group,
	dict: Group,
	}


class ViewCommon:
	''' Common base for Qt's View rendering and Offscreen rendering. It provides common methods to render and interact with a view.
		
		You should always use one of its subclass.
	'''
	def __init__(self, scene, projection=None, navigation=None):
		# interaction methods
		self.projection = projection or globals()[settings.scene['projection']]()
		self.navigation = navigation or globals()[settings.controls['navigation']]()

		# render parameters
		self.scene = scene if isinstance(scene, Scene) else Scene(scene)
		self.uniforms = {'proj':fmat4(1), 'view':fmat4(1), 'projview':fmat4(1)}	# last frame rendering constants
		self.targets = []
		self.steps = []
		self.step = 0
		self.stepi = 0

		# dump targets
		self.map_depth = None
		self.map_idents = None
		self.fresh = set()	# set of refreshed internal variables since the last render

	# -- internal frame system --

	def refreshmaps(self):
		''' Load the rendered frames from the GPU to the CPU

			- When a picture is used to GPU rendering it's called 'frame'
			- When it is dumped to the RAM we call it 'map' in this library
		'''
		if 'fb_ident' not in self.fresh:
			self.makeCurrent()	# set the scene context as current opengl context
			with self.scene.ctx as ctx:
				#ctx.finish()
				self.fb_ident.read_into(self.map_ident, viewport=self.fb_ident.viewport, components=2)
				self.fb_ident.read_into(self.map_depth, viewport=self.fb_ident.viewport, components=1, attachment=-1, dtype='f4')
			self.fresh.add('fb_ident')
			#from PIL import Image
			#Image.fromarray(self.map_ident*16, 'I;16').show()

	def render(self):
		# prepare the view uniforms
		w, h = self.fb_screen.size
		self.uniforms['view'] = view = self.navigation.matrix()
		self.uniforms['proj'] = proj = self.projection.matrix(w/h if h > 0 else 0, self.navigation.distance)
		self.uniforms['projview'] = proj * view
		self.fresh.clear()

		# call the render stack
		self.scene.render(self)

	def identstep(self, nidents):
		''' Updates the amount of rendered idents and return the start ident for the calling rendering pass?
			Method to call during a renderstep
		'''
		s = self.step
		self.step += nidents
		self.steps[self.stepi] = self.step-1
		self.stepi += 1
		return s

	def setup_ident(self):
		# steps for fast fast search of displays with the idents
		self.stepi = 0
		self.step = 1
		if 'ident' in self.scene.stacks and len(self.scene.stacks['ident']) != len(self.steps):
			self.steps = [0] * len(self.scene.stacks['ident'])
		# ident rendering setup
		ctx = self.scene.ctx
		ctx.multisample = False
		ctx.enable_only(mgl.DEPTH_TEST)
		ctx.blend_func = mgl.ONE, mgl.ZERO
		ctx.blend_equation = mgl.FUNC_ADD
		self.target.clear(0)

	def setup_screen(self):
		# screen rendering setup
		ctx = self.scene.ctx
		ctx.multisample = True
		ctx.enable_only(mgl.BLEND | mgl.DEPTH_TEST)
		ctx.blend_func = mgl.SRC_ALPHA, mgl.ONE_MINUS_SRC_ALPHA
		ctx.blend_equation = mgl.FUNC_ADD
		
		background = settings.display['background_color']
		if len(background) == 3:
			self.target.clear(*background, alpha=1)
		elif len(background) == 4:
			self.target.clear(*background)
		else:
			raise ValueError(f"background_color must be a RGB or RGBA tuple, currently {background}")

	def preload(self):
		''' Internal method to load common resources '''
		ctx, resources = self.scene.ctx, self.scene.resources
		resources['shader_ident'] = ctx.program(
					vertex_shader=open(resourcedir+'/shaders/object-ident.vert').read(),
					fragment_shader=open(resourcedir+'/shaders/ident.frag').read(),
					)

		resources['shader_subident'] = ctx.program(
					vertex_shader=open(resourcedir+'/shaders/object-item-ident.vert').read(),
					fragment_shader=open(resourcedir+'/shaders/ident.frag').read(),
					)

	# -- methods to deal with the view --

	def somenear(self, point: ivec2, radius=None) -> ivec2:
		''' Return the closest coordinate to coords, (within the given radius) for which there is an object at
			So if objnear is returning something, objat and ptat will return something at the returned point
		'''
		if radius is None:
			radius = settings.controls['snap_dist']
		self.refreshmaps()
		for x,y in snailaround(point, (self.map_ident.shape[1], self.map_ident.shape[0]), radius):
			ident = int(self.map_ident[-y, x])
			if ident:
				return uvec2(x,y)

	def ptat(self, point: ivec2) -> fvec3:
		''' Return the point of the rendered surfaces that match the given window coordinates '''
		self.refreshmaps()
		viewport = self.fb_ident.viewport
		depthred = float(self.map_depth[-point.y,point.x])
		x =  (point.x/viewport[2] *2 -1)
		y = -(point.y/viewport[3] *2 -1)

		if depthred == 1.0:
			return None
		else:
			view = self.uniforms['view']
			proj = self.uniforms['proj']
			a,b = proj[2][2], proj[3][2]
			depth = b/(depthred + a) * 0.5	# TODO get the true depth  (can't get why there is a strange factor ... opengl trick)
			#near, far = self.projection.limits  or settings.display['view_limits']
			#depth = 2 * near / (far + near - depthred * (far - near))
			#print('depth', depth, depthred)
			return vec3(fvec3(affineInverse(view) * fvec4(
						depth * x /proj[0][0],
						depth * y /proj[1][1],
						-depth,
						1)))

	def ptfrom(self, point: ivec2, center: fvec3) -> fvec3:
		''' 3D point below the cursor in the plane orthogonal to the sight, with center as origin '''
		view = self.uniforms['view']
		proj = self.uniforms['proj']
		viewport = self.fb_ident.viewport
		x =  (point.x/viewport[2] *2 -1)
		y = -(point.y/viewport[3] *2 -1)
		depth = (view * fvec4(fvec3(center),1))[2]
		return vec3(fvec3(affineInverse(view) * fvec4(
					-depth * x /proj[0][0],
					-depth * y /proj[1][1],
					depth,
					1)))

	def itemat(self, point: ivec2) -> 'key':
		''' Return the key path of the object at the given screen position (widget relative).
			If no object is at this exact location, None is returned
		'''
		self.refreshmaps()
		point = uvec2(point)
		ident = int(self.map_ident[-point.y, point.x])
		if ident and 'ident' in self.scene.stacks:
			rdri = bisect(self.steps, ident)
			if rdri == len(self.steps):
				print('internal error: object ident points out of idents list')
			while rdri > 0 and self.steps[rdri-1] == ident:	rdri -= 1
			if rdri > 0:	subi = ident - self.steps[rdri-1] - 1
			else:			subi = ident - 1
			
			if rdri >= len(self.scene.stacks['ident']):
				print('wrong identification index', ident, self.scene.stacks['ident'][-1])
				nprint(self.scene.stacks['ident'])
				return
			
			return (*self.scene.stacks['ident'][rdri][0], subi)

	# -- view stuff --

	def look(self, position: fvec3=None):
		''' Make the scene navigation look at the position.
			This is changing the camera direction, center and distance.
		'''
		if not position:	position = self.scene.box().center
		dir = position - fvec3(affineInverse(self.navigation.matrix())[3])
		if not dot(dir,dir) > 1e-6 or not isfinite(position):	return

		if isinstance(self.navigation, Turntable):
			self.navigation.yaw = atan2(dir.x, dir.y)
			self.navigation.pitch = -atan2(dir.z, length(dir.xy))
			self.navigation.center = position
			self.navigation.distance = length(dir)
		elif isinstance(self.navigation, Orbit):
			focal = self.orient * fvec3(0,0,1)
			self.navigation.orient = quat(dir, focal) * self.navigation.orient
			self.navigation.center = position
			self.navigation.distance = length(dir)
		else:
			raise TypeError("navigation {} is not supported by 'look'".format(type(self.navigation)))

	def adjust(self, box:Box=None):
		''' Make the navigation camera large enough to get the given box in .
			This is changing the zoom level
		'''
		if not box:	box = self.scene.box()
		if box.isempty():	return

		# get the most distant point to the focal axis
		invview = affineInverse(self.navigation.matrix())
		camera, look = fvec3(invview[3]), fvec3(invview[2])
		dist = length(noproject(box.center-camera, look)) + max(glm.abs(box.width))/2 * 1.1
		if not dist > 1e-6:	return

		# adjust navigation distance
		if isinstance(self.projection, Perspective):
			self.navigation.distance = dist / tan(self.projection.fov/2)
		elif isinstance(self.projection, Orthographic):
			self.navigation.distance = dist / self.projection.size
		else:
			raise TypeError('projection {} not supported'.format(type(self.projection)))

	def center(self, center: fvec3=None):
		''' Relocate the navigation to the given position .
			This is translating the camera.
		'''
		if not center:	center = self.scene.box().center
		if not isfinite(center):	return

		self.navigation.center = center


class Offscreen(ViewCommon):
	''' Object allowing to perform offscreen rendering, navigate and get information from screen as for a normal window 
	'''
	def __init__(self, scene, size=uvec2(400,400), projection=None, navigation=None):
		global global_context
		
		super().__init__(scene, projection=projection, navigation=navigation)
		
		if global_context:
			self.scene.ctx = global_context
		else:
			self.scene.ctx = global_context = mgl.create_standalone_context(requires=opengl_version)
		self.scene.ctx.line_width = settings.display["line_width"]

		self.init(size)
		self.preload()

	def init(self, size):
		w, h = size

		ctx = self.scene.ctx
		assert ctx, 'context is not initialized'

		# self.fb_frame is already created and sized by Qt
		self.fb_screen = ctx.simple_framebuffer(size)
		self.fb_ident = ctx.simple_framebuffer(size, components=3, dtype='f1')
		self.targets = [ ('screen', self.fb_screen, self.setup_screen),
						 ('ident', self.fb_ident, self.setup_ident)]
		self.map_ident = np.empty((h,w), dtype='u2')
		self.map_depth = np.empty((h,w), dtype='f4')
		
	@property
	def size(self):		
		return self.fb_screen.size
		
	def width(self):
		return self.fb_screen.size[0]
		
	def height(self):
		return self.fb_screen.size[1]
	
	def resize(self, size):
		if size != self.fb_screen.size:
			self.ctx.finish()
			self.init(size)

	def render(self):
		super().render()
		return Image.frombytes('RGBA', tuple(self.size), self.fb_screen.read(components=4), 'raw', 'RGBA', 0, -1)


class View(ViewCommon, QOpenGLWidget):
	''' Qt widget to render and interact with displayable objects.
		It holds a scene as renderpipeline.

		Attributes:

			scene:        the `Scene` object displayed
			projection:   `Perspective` or `Orthographic`
			navigation:   `Orbit` or `Turntable`
			tool:         list of callables in priority order to receive events

			targets:     render targets matching those requested in `scene.stacks`
			uniforms:    parameters for rendering, used in shaders
	'''
	def __init__(self, scene, projection=None, navigation=None, parent=None):
		# super init
		QOpenGLWidget.__init__(self, parent)
		self.setFormat(qt_surface_format())
		
		# ugly trick to receive interaction events in a different function than QOpenGLWidget.event (that one is locking the GIL during the whole rendering, killing any possibility of having a computing thread aside)
		# that event reception should be in the current widget ...
		self.handler = GhostWidget(self)
		self.handler.setFocusPolicy(Qt.StrongFocus)
		self.handler.setAttribute(Qt.WA_AcceptTouchEvents, True)
		self.setFocusProxy(self.handler)

		ViewCommon.__init__(self, scene, projection=projection, navigation=navigation)
		self.tool = [Tool(self.navigation.tool, self)] # tool stack, the last tool is used for input events, until it is removed 

	def init(self):
		w, h = self.width(), self.height()

		ctx = self.scene.ctx
		assert ctx, 'context is not initialized'

		# self.fb_screen is already created and sized by Qt
		self.fb_screen = ctx.detect_framebuffer(self.defaultFramebufferObject())
		self.fb_ident = ctx.simple_framebuffer((w, h), components=3, dtype='f1')
		self.targets = [ ('screen', self.fb_screen, self.setup_screen),
						 ('ident', self.fb_ident, self.setup_ident)]
		self.map_ident = np.empty((h,w), dtype='u2')
		self.map_depth = np.empty((h,w), dtype='f4')

	def render(self):
		# set the opengl current context from Qt (doing it only from moderngl interferes with Qt)
		self.makeCurrent()
		ViewCommon.render(self)

	
	# -- view stuff --

	def look(self, position: fvec3=None):
		ViewCommon.look(self, position)
		self.update()

	def adjust(self, box:Box=None):
		ViewCommon.adjust(self, box)
		self.update()

	def center(self, center: fvec3=None):
		ViewCommon.center(self, center)
		self.update()
	
	def somenear(self, point: QPoint, radius=None) -> QPoint:
		some = ViewCommon.somenear(self, qt_2_glm(point), radius)
		if some:
			return glm_to_qt(some)

	def ptat(self, point: QPoint) -> fvec3:
		return ViewCommon.ptat(self, qt_2_glm(point))

	def ptfrom(self, point: QPoint, center: fvec3) -> fvec3:
		return ViewCommon.ptfrom(self, qt_2_glm(point), center)

	def itemat(self, point: QPoint) -> 'key':
		return ViewCommon.itemat(self, qt_2_glm(point))
	

	# -- event system --

	def inputEvent(self, evt):
		''' Default handler for every input event (mouse move, press, release, keyboard, ...)
			When the event is not accepted, the usual matching Qt handlers are used (mousePressEvent, KeyPressEvent, etc).

			This function can be overwritten to change the view widget behavior.
		'''
		# send the event to the current tools using the view
		if self.tool:
			for tool in reversed(self.tool):
				tool(evt)
				if evt.isAccepted():	return

		# send the event to the scene objects, descending the item tree
		if isinstance(evt, QMouseEvent) and evt.type() in (QEvent.MouseButtonPress, QEvent.MouseButtonRelease, QEvent.MouseButtonDblClick, QEvent.MouseMove):
			pos = self.somenear(evt.pos())
			if pos:
				key = self.itemat(pos)
				if key:
					self.control(key, evt)
					if evt.isAccepted():	return

			# if clicks are not accepted, then some following keyboard events may not come to the widget
			# NOTE this also discarding the ability to move the window from empty areas
			if evt.type() == QEvent.MouseButtonPress:
				evt.accept()

	def control(self, key, evt):
		''' Transmit a control event successively to all the displays matching the key path stages.
			At each level, if the event is not accepted, it transmits to sub items

			This function can be overwritten to change the interaction with the scene objects.
		'''
		disp = self.scene.displays
		stack = []
		for i in range(1,len(key)):
			disp = disp[key[i-1]]
			disp.control(self, key[:i], key[i:], evt)
			if evt.isAccepted(): return
			stack.append(disp)

		if evt.type() == QEvent.MouseButtonPress and evt.button() == Qt.LeftButton:
			disp = stack[-1]
			# select what is under cursor
			if type(disp).__name__ in ('SolidDisplay', 'WebDisplay'):
				disp.vertices.selectsub(key[-1])
				disp.selected = any(disp.vertices.flags & 0x1)
			else:
				disp.selected = not disp.selected
			# make sure that a display is selected if one of its sub displays is
			for disp in reversed(stack):
				if hasattr(disp, '__iter__'):
					disp.selected = any(sub.selected	for sub in disp)
			self.update()

	# -- Qt things --

	def initializeGL(self):
		# retrieve global shared context if available
		global global_context
		
		if QApplication.testAttribute(Qt.AA_ShareOpenGLContexts):
			if not global_context:
				# global_context = mgl.create_context()
				global_context = mgl.get_context()
			self.scene.ctx = global_context
		# or create a context
		else:
			# self.scene.ctx = mgl.create_context()
			self.scene.ctx = mgl.get_context()
		self.init()
		self.preload()

	def paintGL(self):
		self.makeCurrent()
		self.render()

	def resizeEvent(self, evt):
		QOpenGLWidget.resizeEvent(self, evt)
		self.handler.resize(self.size())
		self.init()
		self.update()

	def changeEvent(self, evt):
		# detect theme change
		if evt.type() == QEvent.PaletteChange and settings.display['system_theme']:
			settings.use_qt_colors()
		return QOpenGLWidget.changeEvent(self, evt)

class GhostWidget(QWidget):
	def __init__(self, parent):
		super().__init__(parent)
		
	def event(self, evt):
		if isinstance(evt, QInputEvent):
			evt.ignore()
			self.parent().inputEvent(evt)
			if evt.isAccepted():	return True
		elif isinstance(evt, QFocusEvent):
			self.parent().event(evt)
		return super().event(evt)


def snail(radius):
	''' Generator of coordinates snailing around 0,0 '''
	x = 0
	y = 0
	for r in range(radius):
		for x in range(-r,r):		yield ivec2(x,-r)
		for y in range(-r,r):		yield ivec2(r, y)
		for x in reversed(range(-r,r)):	yield ivec2(x, r)
		for y in reversed(range(-r,r)):	yield ivec2(-r,y)

def snailaround(pt, box, radius):
	''' Generator of coordinates snailing around pt, coordinates that goes out of the box are skipped '''
	cx,cy = pt
	mx,my = box
	for rx,ry in snail(radius):
		x,y = cx+rx, cy+ry
		if 0 <= x and x < mx and 0 <= y and y < my:
			yield ivec2(x,y)

def glm_to_qt(p): 	return QPoint(p.x, p.y)
def qt_2_glm(p):	return ivec2(p.x(), p.y())

'''
		-- generators helpers --
'''

class Generated(object):
	''' Generator that has a returned value '''
	__slots__ = 'generator', 'value'
	def __init__(self, generator):	self.generator = generator
	def __iter__(self):				self.value = yield from self.generator

class Dispatcher(object):
	''' Iterable object that holds a generator built by passing self as first argument
		it allows the generator code to dispatch references to self.
		NOTE:  at contrary to current generators, the code before the first yield is called at initialization
	'''
	__slots__ = 'generator', 'value'
	def __init__(self, func=None, *args, **kwargs):
		self.func = func
		self.generator = self._run(func, *args, **kwargs)
		# run the generator until the first yield
		next(self.generator, None)
	def _run(self, func, *args, **kwargs):
		self.value = yield from func(self, *args, **kwargs)
		
	def __repr__(self):
		return '<{} on {}>'.format(type(self).__name__, self.func)
		
	def send(self, value):	return self.generator.send(value)
	def __iter__(self):		return self.generator
	def __next__(self):		return next(self.generator)

class Tool(Dispatcher):
	''' Generator wrapping an yielding function, that unregisters from view.tool once the generator is over '''
	def _run(self, func, *args, **kwargs):
		try:	
			self.value = yield from func(self, *args, **kwargs)
		except StopTool:
			pass
		try:	
			args[0].tool.remove(self)
		except ValueError:	
			pass
	
	def __call__(self, evt):
		try:	return self.send(evt)
		except StopIteration:	pass
		
	def stop(self):
		if self.generator:
			try:	self.generator.throw(StopTool())
			except StopTool:	pass
			except StopIteration:	pass
			self.generator = None
	def __del__(self):
		self.stop()
	
class StopTool(Exception):
	''' Used to stop a tool execution '''
	pass


# temporary examples
if False:

	def tutu(self, main):
		evt = yield
		gnagna
		scene.tool = self.send
		budu.iterator = self

	Tool(tutu, main)
	



'''
	Module exposing many functions to create/visualize standard parts.
	Those are following the ISO (metric) specifications as often as possible.
	
	Most parts are as easy to create as:
	
		>>> nut(8)    # nut with nominal diameter 8mm
		>>> screw(8, 16)   # screw with match diameter 8mm, and length 16mm
		
	The parts usually have many optional parameters that default to usual recommended values. You can override this by setting the keyword arguments:
	
		>>> screw(8, 16, head='button', drive='slot')
'''

from operator import itemgetter
from itertools import accumulate

from .mathutils import *
from .primitives import *
from .kinematic import Solid
from .joints import Revolute
from .mesh import Mesh, Web, Wire, web, wire
from .generation import *
from .offseting import *
from .blending import *
from .boolean import pierce, union, difference, intersection
from .bevel import *
from .io import cachefunc

#cachefunc = lambda x:x	# debug purpose


__all__ = [	'nut', 'screw', 'washer', 'bolt',
			'coilspring_compression', 'coilspring_tension', 'coilspring_torsion',
			'bearing', 'slidebearing',
			'section_s', 'section_w', 'section_c', 'section_l', 'section_tslot', 'grooves',
			'screw_slot', 'bolt_slot', 'bearing_slot_exterior', 'bearing_slot_interior',
			'circular_screwing',
			'grooves_profile', 'grooves',
			'stfloor', 'stceil',
			]

			
# --------- numeric stuff --------------------

standard_digits = [.0, 1., .5, .4, .6, .8, .2, .25, .75]

def stfloor(x, precision=0.1):
	''' Return a numeric value fitting x, with lower digits being the under closest digits from `standard_digits` 
		
		`precision` gives the relative tolerance interval for the returned number, so we are sure it lays in `[x*(1-precision), x]`
	'''
	if hasattr(x, '__getitem__'):
		return type(x)(stfloor(e)  for e in x)
	if not isfinite(x):
		return x
	s, x = sign(x), abs(x)
	
	base = 10
	magnitude = base**floor(log(x) / log(base))
	ratio = x / magnitude
	
	for j in range(2 + int(ceil(-log(precision) / log(base)))):
		for st in standard_digits:
			candidate = floor(ratio) + st
			if candidate > ratio + 2*NUMPREC:
				continue
			if ratio*(1-precision) + NUMPREC <= candidate:
				return candidate * magnitude * s
		magnitude /= base
		ratio *= base
	raise RuntimeError('failed to converge, this is a bug')

def stceil(x, precision=0.1):
	''' Return a numeric value fitting x, with lower digits being the above closest digits from `standard_digits` 
		
		`precision` gives the relative tolerance interval for the returned number, so we are sure it lays in `[x, x*(1+precision)]`
	'''
	if hasattr(x, '__getitem__'):
		return type(x)(stceil(e)  for e in x)
	if not isfinite(x):
		return x
	s, x = sign(x), abs(x)
	
	base = 10
	magnitude = base**ceil(log(x) / log(base))
	ratio = x / magnitude
	
	for j in range(2 + int(ceil(-log(precision) / log(base)))):
		for st in standard_digits:
			candidate = floor(ratio) + st
			if candidate < ratio - 2*NUMPREC:
				continue
			if ratio*(1+precision) + NUMPREC >= candidate:
				return candidate * magnitude * s
		magnitude /= base
		ratio *= base
	raise RuntimeError('failed to converge, this is a bug')

# --------- screw stuff -----------------------
	
@cachefunc
def screw(d, length, filet_length=None, head='SH', drive=None, detail=False):
	''' Create a standard screw using the given drive and head shapes
	
	Parameters:
		d:             nominal diameter of the screw
		length:        length from the screw head to the tip of the screw
		filet_length:  length of the filet on the screw (from the tip), defaults to `length`
		
		head:          name of the screw head shape
		drive:			name of the screw drive shape
		detail:   if True, the thread surface will be generated, else it will only be a cylinder
		
	It's also possible to specify head and drive at once, using their codenames:
	
		>>> screw(5, 16, head='SHT')   # socket head and torx drive
		
	Available screw heads:
		* socket (default)
		* button
		* flat (aka. cone)
		
	Available screw drives:
		* hex
		* torx  *(not yet available)*
		* phillips (cross)  *(not yet available)*
		* slot
	
	All possible shapes:
		* see wikipedia for the [drive types](https://en.wikipedia.org/wiki/List_of_screw_drives)
		* see [here for a guide of what to use](https://www.homestratosphere.com/types-of-screws/)
		* see wikipedia for [standard screw thread](https://en.wikipedia.org/wiki/ISO_metric_screw_thread)
		
		
	'''
	if filet_length is None:	filet_length = length - 0.05*d
	elif length < filet_length:	raise ValueError('filet_length must be smaller than length')
	
	head, drive = screw_spec(head, drive)
	head = globals()['screwhead_'+head](d)
	if drive:
		drive = globals()['screwdrive_'+drive](d) .transform(boundingbox(head).max[2]*Z)
		head = intersection(head, drive)
		
	r = 0.5*d
	axis = Axis(O, Z, interval=(-length,r))
	top = head.box().min.z
	body = revolution(wire([
					vec3(r, 0, top),
					vec3(r, 0, min(top, -length+filet_length)),
					vec3(r, 0, -length+r*0.2),
					vec3(r*0.8, 0, -length),
					vec3(0, 0, -length),
					]) .segmented())
	screw = (body + head).finish().option(color=settings.colors['bolt'])
	return Solid(part=screw, axis=axis)

def screwdrive_torx(d):
	indev
	
def screwdrive_hex(d):
	base = regon((-0.3*d*Z, -Z), 0.5*d, 6)
	socket = extrusion(base, d*Z) + blendloop(base, center=-0.6*d*Z, weight=-1)
	socket.mergeclose()
	return socket
	
def screwdrive_cross(d):
	indev
	
def screwdrive_slot(d):
	w = 0.15*d
	h = 0.3*d
	e = 2*d
	return extrusion(web([
				vec3(-e, w, h),
				vec3(-e, w, -h),
				vec3(-e, -w, -h),
				vec3(-e, -w, h),
				]) .segmented(), 
				2*e*X)

def screwhead_socket(d):
	''' Screw head shape for socket head (SH) '''
	r = h = 0.7*d
	c = 0.05*d
	
	profile = wire([
			vec3(d/2,	0,	-c),
			vec3(d/2+c,	0,	0),
			vec3(r,     0,  0), 
			vec3(r,     0,  h-c),
			vec3(r-c,   0,  h),
			vec3(0,     0,  h),
			]) .segmented()
	head = revolution(profile.flip())
	head.finish()
	return head
	
def screwhead_hex(d):
	''' Screw head shape for hex head (HH) '''
	r = 0.9*d
	h = 0.6*d
	c = 0.05*d
	
	profile = extrusion(regon((-d*Z,Z), r, 6), 2*d*Z)
	cone = revolution(web([
			vec3(0,       0, h),
			vec3(0.8*r,   0, h),
			vec3(1.01*r,  0, h*0.8),
			vec3(1.01*r,  0, 0),
			vec3(0.5*d+c, 0, 0),
			vec3(0.5*d,   0, -c),
			]) .segmented())
	head = intersection(cone, profile)
	head.finish()
	return head
	
def screwhead_button(d):
	r = 0.95*d
	h = 0.5*d
	c = 0.05*d
	
	profile = [
		wire([
			vec3(0, 0, h), 
			vec3(0, 0.5*d, h),
			]),
		TangentEllipsis(
			vec3(0, 0.5*d, h), 
			vec3(0, 0.9*r, h),
			vec3(0, r, 0.1*d)),
		wire([
			vec3(0, r, 0.1*d),
			vec3(0, r, 0),
			vec3(0, 0.5*d+c, 0),
			vec3(0, 0.5*d, -c),
			]) .segmented(),
		]
	return revolution(profile)
	
def screwhead_flat(d):
	r = d
	h = 0.5*d
	e = 0.1*d
	
	return revolution(web([
		vec3(0, 0, 0),
		vec3(0, r, 0),
		vec3(0, r, -e),
		vec3(0, 0.5*d, -0.5*d-e),
		]) .segmented() ).finish()
	
def screwhead_none(d):
	indev

	
'''
	head shapes:
	Abbreviation 	Expansion 	Comment
	BH 		button head 
	FH 		flat head 	
	OH 		oval head 	
	PH 		Phillips head 	
	RH 		round head 		
	FHP		flat head Phillips 	
	RHP 	round head Phillips 	
	SH 		socket head			Although "socket head" could logically refer to almost any female drive, it refers by convention to hex socket head unless further specified.
	SS 		set screw			The abbreviation "SS" more often means stainless steel. Therefore, "SS cap screw" means "stainless steel cap screw" but "SHSS" means "socket head set screw". As with many abbreviations, users rely on context to diminish the ambiguity, although this reliance does not eliminate it.
	VH		conic head
	
	CS 		cap screw 	
	MS 		machine screw 	
	
	BHCS 	button head cap screw 	
	BHMS 	button head machine screw 	
	FHCS 	flat head cap screw 
	FHSCS 	flat head socket cap screw 	
	FHPMS 	flat head Phillips machine screw 	
	FT 		full thread 	
	HHCS 	hex head cap screw 	
	HSHCS 	Hexalobular socket head cap screws 	
	RHMS 	round head machine screw 	
	RHPMS 	round head Phillips machine screw 	
	SBHCS 	socket button head cap screw 	
	SBHMS 	socket button head machine screw 	
	SHCS 	socket head cap screw 	
	SHSS 	socket head set screw 	Sometimes Socket Head Shoulder Screw.
	STS 	self-tapping screw
	
	[standard screw thread](https://en.wikipedia.org/wiki/ISO_metric_screw_thread)
'''

screwheads_codes = {
	'SH': 'socket',
	'HH': 'hex',
	'VH': 'flat',
	'cone': 'flat',
	'BH': 'button',
	'SBH': 'button',
	'SS': 'none',
	'FT': 'none',
	}
	
screwdrives_codes = {
	'TX': 'torx',
	'T': 'torx',
	'TR': 'torx',
	'CS': 'cap',
	'phillips': 'cross',
	'P': 'cross',
	'PH': 'cross',
	'PH': 'cross',
	}

def screw_spec(head, drive=None):
	if not drive:
		if head.isupper():
			for code in screwdrives_codes:
				if head.endswith(code):
					head, drive = head[:-len(code)], code
					break
	
	head = screwheads_codes.get(head, head)
	drive = screwdrives_codes.get(drive, drive)
	# special cases
	if head != 'hex' and not drive:
		drive = 'hex'
		
	return head, drive


# ------------------- nut stuff ----------------------

@cachefunc
def nut(d, type='hex', detail=False) -> Mesh:
	''' Create a standard nut model using the given shape type 
	
		Parameters:
			d:        nominal diameter of the matching screw
			type:     the nut shape
			detail:   if True, the thread surface will be generated, else it will only be a cylinder
	
		If `d` alone is given, the other parameters default to the ISO specs: https://www.engineersedge.com/iso_flat_washer.htm
		
		Currently only shape 'hex' is available.
	'''
	args = standard_hexnuts[bisect(standard_hexnuts, d, key=itemgetter(0))]
	if args[0] != d:
		raise ValueError('no standard nut for the given diameter')
	return hexnut(*args)

	
def hexnut(d, w, h):
	''' Create an hexagon nut with custom dimensions '''
	# revolution profile
	w *= 0.5
	r = 1.01 * w/cos(radians(30))
	profile = wire([
		vec3(0.5*d,	0,	0.5*h),
		vec3(0.95*w,	0,	0.5*h),
		vec3(r,	0,	0.5*h - (r-w)),
		vec3(r,	0,	-0.5*h + (r-w)),
		vec3(0.95*w,	0,	-0.5*h),
		vec3(0.5*d,	0,	-0.5*h),
		]) .close() .segmented()
	base = revolution(web(profile))
	
	# exterior hexagon shape
	hexagon = regon((-h*Z,Z),  w/cos(radians(30)), 6)
	ext = extrusion(hexagon, 2*h*Z)
	
	# intersect everything
	nut = intersection(base, ext)
	chamfer(nut, nut.frontiers(4,5) + nut.frontiers(0,5), width=d*0.1)

	nut.finish().option(color=settings.colors['bolt'])
	return Solid(
				part=nut, 
				bottom=Axis(-0.5*h*Z, -Z, interval=(0,h)), 
				top=Axis(0.5*h*Z, Z, interval=(0,h)),
				)

''' Iso hexagon nuts according to [EN ISO 4032](https://www.fasteners.eu/standards/ISO/4032/)
	columns:
	* thread
	* w
	* h
'''
standard_hexnuts = [
	(1.6,  3.2, 1.3),
	(2,    4,   1.6),
	(2.5,  5,   2),
	(3,    5.5, 2.4),
	(3.5,  6,   2.8), # non-prefered
	(4,    7,   3.2),
	(5,    8,   4.7),
	(6,    10,  5.2),
	(8,    13,  6.8),
	(10,   16,  8.4),
	(12,   18,  10.8),
	(14,   21,  12.8), # non-prefered
	(16,   24,  14.8),
	(18,   27,  15.8), # non-prefered
	(20,   30,  18),
	(22,   34,  19.4), # non-prefered
	(24,   36,  21.5),
	(27,   41,  23.8), # non-prefered
	(30,   46,  25.6),
	(33,   50,  38.7), # non-prefered
	(36,   55,  31),
	(39,   60,  33.4), # non-prefered
	(42,   65,  34),
	(45,   70,  36), # non-prefered
	(48,   75,  38),
	(52,   80,  42), # non-prefered
	(56,   85,  45),
	(60,   90,  48), # non-prefered
	(64,   95,  51),
	]
	
	
# -------------- washer stuff ----------------------
	
@cachefunc
def washer(d, e=None, h=None) -> Mesh:
	''' Create a standard washer.
		Washers are useful to offset screws and avoid them to scratch the mount part
		
		Parameters:
			d:        the nominal interior diameter (screw or anything else),
			           the exact washer interior is slightly bigger
			e:        exterior diameter
			h:        height/thickness
			
		If `d` alone is given, the other parameters default to the ISO specs: https://www.engineersedge.com/iso_flat_washer.htm
	'''
	if e is None and h is None:
		args = standard_washers[bisect(standard_washers, d, key=itemgetter(0))]
		if abs(args[0] - d)/d < 0.2:
			_, d, e, h = args
		else:
			raise ValueError('no standard nut for the given diameter')
	else:
		d *= 1.1
		if e is None:	e = d*2
		if h is None:	h = d*0.1
	
	return Solid(
				part=thicken(revolution(web([e/2*X, d/2*X])), h).option(color=settings.colors['bolt']), 
				top=Axis(h*Z, Z, interval=(0,h)), 
				bottom=Axis(O, -Z, interval=(0,h)),
				)


''' Metric washers according to https://www.engineersedge.com/iso_flat_washer.htm
	columns;
	* nominal screw
	* interior size
	* exterior size
	* thickness
'''
standard_washers	= [
	(1.6, 1.7, 4,   0.3),
	(2,   2.2, 5,   0.3),
	(2.5, 2.7, 6,   0.5),
	(2.6, 2.8, 7,   0.5),
	(3,   3.2, 7,   0.5),
	(3.5, 3.7, 8,   0.5),
	(4,   4.3, 9,   0.8),
	(5,   5.3, 10,  1),
	(6,   6.4, 12,  1.6),
	(7,   7.4, 14,  1.6),
	(8,   8.4, 16,  1.6),
	(10,  10.5, 20, 2),
	(12,  13,  24,  2.5),
	(14,  15,  28,  2.5),
	(16,  17,  30,  3),
	(18,  19,  34,  3),
	(20,  21,  37,  3),
	(22,  23,  39,  3),
	(24,  25,  44,  4),
	(27,  28,  50,  4),
	(30,  31,  56,  4),
	(33,  34,  60,  5),
	(36,  37,  66,  5),
	(39,  40,  72,  6),
	(42,  43,  78,  7),
	(45,  46,  85,  7),
	(48,  50,  92,  8),
	(52,  54,  98,  8),
	(56,  58,  105, 9),
	]
	
	
def section_s(height=1, width=None, flange=None, thickness=None) -> Web:
	''' Standard S (short flange) section. Very efficient to support flexion efforts.
	'''
	if width is None:	width = 0.4 * height
	if flange is None:	flange = 0.036 * height
	if thickness is None:	thickness = 0.054 * height
	assert width > 3*thickness+2*flange and height > 2*flange+2*thickness
	
	base = wire([
		vec3(width/2, height/2, 0),
		vec3(width/2, height/2-flange, 0),
		vec3(thickness/2, height/2-flange-0.1*(width/2-thickness/2), 0),
		]) .flip()
	base = base + base.transform(scaledir(X,-1)).flip()
	base = base + base.transform(scaledir(Y,-1)).flip()
	section = web(base.close().segmented())
	filet(section, section.frontiers(0,5,10,4,6,11), radius=thickness*0.4, resolution=('div',2))
	filet(section, section.frontiers(1,0,4,3,6,7,10,9), radius=flange, resolution=('div',2))
	
	#notes = [
		#note_distance(base.points[0], base.points[0]*vec3(1,-1,1), offset=2*flange*X),
		#note_distance(base.points[0], base.points[1], offset=flange*X),
		#note_distance(base.points[0], base.points[0]*vec3(-1,1,1), offset=flange*Y),
		#note_distance(thickness/2*X, -thickness/2*X),
		#]
	
	return section.finish()

def section_w(height=1, width=None, flange=None, thickness=None) -> Web:
	''' Standard W (wide flange) section. It is slightly different than a S section in that the flanges are straight and are usally wider.
	'''
	if width is None:	width = 0.6 * height
	if flange is None:	flange = 0.036 * height
	if thickness is None:	thickness = 0.054 * height
	assert width > 3*thickness+2*flange and height > 2*flange+2*thickness
	
	base = wire([
		vec3(width/2, height/2, 0),
		vec3(width/2, height/2-flange, 0),
		vec3(thickness/2, height/2-flange, 0),
		]) .flip()
	base = (base + base.transform(scaledir(X,-1)).flip() ).segmented()
	base = base + base.transform(scaledir(Y,-1)).flip()
	base.close()
	section = web(base)
	filet(section, section.frontiers(0,5,10,4,6,11), radius=thickness*0.4, resolution=('div',2))
	filet(section, section.frontiers(1,0,4,3,6,7,10,9), radius=flange*0.5, resolution=('div',2))
	return section.finish()
	
def section_l(a=1, b=None, thickness=None) -> Wire:
	''' Standard L section '''
	if b is None:	b = a
	if thickness is None:	thickness = 0.05*max(a,b)
	assert a > 3*thickness and b > 3*thickness

	section = wire([
		vec3(0, 0, 0),
		vec3(0, a, 0),
		vec3(thickness, a, 0),
		vec3(thickness, thickness, 0),
		vec3(b, thickness, 0),
		vec3(b, 0, 0),
		]) .close() .segmented() .flip()

	filet(section, [2,3,4], radius=thickness*0.8, resolution=('div',2))
	return section.finish()
	
def section_c(height=1, width=None, thickness=None) -> Web:
	''' Standard C section '''
	if width is None:	width = 0.6*height
	if thickness is None:	thickness = 0.05*height
	assert width > 3*thickness
	
	base = wire([
		vec3(0, height/2, 0),
		vec3(width, height/2, 0),
		vec3(width, height/2-thickness, 0),
		vec3(thickness, height/2-thickness, 0),
		]) .flip()
	base = base + base.transform(scaledir(Y,-1)).flip()
	section = web(base.close().segmented())
	
	filet(section, section.frontiers(0,1,5,7,6), radius=0.8*thickness, resolution=('div',2))
	return section.finish()

def section_tslot(size=1, slot=None, thickness=None, depth=None) -> Web:
	''' Standard T-Slot section. That section features slots on each side to put nuts at any position.
	'''
	if slot is None:	slot = 0.3*size
	if thickness is None:	thickness = 0.08*size
	if depth is None:	depth = 0.2*size
	b = depth-thickness
	
	center = Circle((O,-Z), size/2-depth-2*thickness)
	base = wire([
		vec3(size/2, size/2, 0),
		vec3(size/2, slot/2+thickness, 0),
		vec3(size/2-thickness, slot/2, 0),
		vec3(size/2-thickness, slot/2+b, 0),
		vec3(size/2-thickness-depth+b, slot/2+b, 0),
		vec3(size/2-thickness-depth, slot/2, 0),
		]) .flip()
	base = (base + base.transform(scaledir(normalize(vec3(1,-1,0)), -1)) .flip() ).segmented()
	base = base + base.transform(scaledir(X, -1)) .flip()
	base = base + base.transform(scaledir(Y, -1)) .flip()
	section = web([
		center,
		base.close().segmented(),
		])
	filet(section, section.frontiers(5,7, 41,43, 31,29, 17,19), radius=thickness/2, resolution=('div',2))
	
	#notes = [
		#note_distance(base.points[2], base.points[5], offset=-c/2*Y),
		#note_distance(base.points[1], base.points[2], offset=-c/2*Y, project=X),
		#note_distance(base.points[2], base.points[2]*vec3(1,-1,1), offset=s*X),
		#note_distance(base.points[0], base.points[0]*vec3(1,-1,1), offset=2*s*X),
		#]
	
	return section.finish()
	
''' Standard IPN sections (S sections)
columns:
	- h
	- b
	- s
	- t
'''
standard_ipn = [
	(80, 42, 3.9, 5.9),
	(100, 50, 4.5, 6.8),
	(120, 58, 5.1, 7.7),
	(140, 66, 5.7, 8.6),
	(160, 74, 6.3, 9.5),
	(180, 82, 6.9, 10.4),
	(200, 90, 7.5, 11.3),
	(220, 98, 8.1, 12.2),
	(240, 106, 8.7, 13.1),
	(260, 113, 9.4, 14.1),
	(280, 119, 10.1, 15.2),
	(300, 125, 10.8, 16.2),
	(320, 131, 11.5, 17.3),
	(340, 137, 12.2, 18.3),
	(360, 143, 13.0, 19.5),
	(380, 149, 13.7, 20.5),
	(400, 155, 14.4, 21.6),
	(450, 170, 16.2, 24.3),
	(500, 185, 18.0, 27.0),
	(550, 200, 19.0, 30.0),
	(600, 215, 21.6, 32.4),
	]

	
# --------------------- coilspring stuff ------------------------

@cachefunc
def coilspring_compression(length, d=None, thickness=None, solid=True):
	''' Return a Mesh model of a croilspring meant for use in compression
	
		Parameters:
			length:     the distance between its two ends
			d:          the exterior diameter (the coilspring can fit in a cylinder of that diameter)
			thickness:  the wire diameter of the spring (useless if solid is disabled)
			solid:      disable it to get only the tube path of the coil, and have a `Wire` as return value
	'''
	if not d:			d = length*0.2
	if not thickness:	thickness = d*0.1
	r = d/2 - thickness		# coil radius
	e = r					# coil step
	div = settings.curve_resolution(d*pi, 2*pi)
	step = 2*pi/(div+1)
	
	t = 0
	
	t0, z0 = t, -0.5*length
	top = []
	for t in linrange(t0, t0 + 4*pi, step):
		top.append( vec3(r*cos(t), r*sin(t), z0 + (t-t0)/(2*pi) * thickness) )
	
	t0, z0 = t, -0.5*length + 2*thickness
	coil = []
	for t in linrange(t0, t0 + 2*pi * (length-4*thickness) / e, step):
		coil.append( vec3(r*cos(t), r*sin(t), z0 + (t-t0)/(2*pi) * e) )
	
	t0, z0 = t, 0.5*length - 2*thickness
	bot = []
	for t in linrange(t0, t0 + 4*pi, step):
		bot.append( vec3(r*cos(t), r*sin(t), z0 + (t-t0)/(2*pi) * thickness) )
		
	path = Wire(top) + Wire(coil).qualify('spring') + Wire(bot)
	
	if not solid:
		return path
	
	return Solid(
			part=tube(
				flatsurface(Circle(
					(path[0],Y), 
					thickness/2, 
					resolution=('div',6)
					)) .flip(), 
				path,
				),
			axis=Axis(O, Z, interval=(-length/2, length/2)),
			top=0.5*length*Z,
			bottom=-0.5*length*Z,
			)
	
@cachefunc
def coilspring_tension(length, d=None, thickness=None, solid=True):
	''' Return a Mesh model of a croilspring meant for use in tension 
	
		Parameters:
			length:     the distance between its two hooks
			d:          the exterior diameter (the coilspring can fit in a cylinder of that diameter)
			thickness:  the wire diameter of the spring (useless if solid is disabled)
			solid:      disable it to get only the tube path of the coil, and have a `Wire` as return value
	'''
	if not d:			d = length*0.2
	if not thickness:	thickness = d*0.1
	r = d/2 - thickness		# coil radius
	e = r					# coil step
	
	# separate the coilspring in 3 parts:  the coil and the 2 hooks at both ides
	spring_length = length - 2*r
	ncoil = floor(spring_length / thickness) - 0.5
	hold = 0.5*length - r
	
	# create coil
	div = settings.curve_resolution(d*pi, 2*pi)
	step = 2*pi/(div+1)
	z0 = -0.5 * ncoil * thickness
	coil = Wire([	vec3(r*cos(t), r*sin(t), z0 + t/(2*pi) * thickness)
					for t in linrange(0, 2*pi * ncoil, step) ]) .qualify('spring')
	# create path with hooks
	path = wire([
				ArcCentered((-0.5*length*Z, X), vec3(0, -r, -0.5*length), -hold*Z),
				ArcThrough(-hold*Z, (-hold*Z + coil[0])*0.5 - 0.5*r*Y, coil[0]),
				coil,
				ArcThrough(coil[-1], (hold*Z + coil[-1])*0.5 - 0.5*r*Y, hold*Z),
				ArcCentered((0.5*length*Z, X), hold*Z, vec3(0, -r, 0.5*length)),
				])
	if not solid:
		return path
	
	return Solid(
			part=tube(
				flatsurface(Circle(
					(path[0],Z), 
					thickness/2, 
					resolution=('div',6)
					)), 
				path,
				),
			axis=Axis(O,Z, interval=(-length/2, length/2)),
			top=0.5*length*Z,
			bottom=-0.5*length*Z,
			)
	
@cachefunc
def coilspring_torsion(arm, 
			angle=radians(45), 
			d=None, 
			length=None, 
			thickness=None, 
			hook=None, 
			solid=True) -> Solid:
	''' Return a Mesh model of a croilspring meant for use in torsion
	
		Parameters:
			arm:        the arms length from the coil axis
			length:     the coil length (and distance between its hooks)
			d:          the exterior diameter (the coilspring can fit in a cylinder of that diameter)
			thickness:  the wire diameter of the spring (useless if solid is disabled)
			hook:       the length of arm hooks (negative for hooks toward the interior)
			solid:      disable it to get only the tube path of the coil, and have a `Wire` as return value
	'''
	if not length:		length = arm*0.5
	if not d:			d = arm
	if not thickness:	thickness = d*0.1
	if not hook:		hook = -length
	r = d/2 - thickness		# coil radius
	e = r					# coil step
	angle = pi - angle
	
	# separate the coilspring in 3 parts:  the coil and the 2 hooks at both ides
	ncoil = ceil(length / thickness) + angle/(2*pi)
	
	# create coil
	div = settings.curve_resolution(d*pi, 2*pi)
	step = 2*pi/(div+1)
	z0 = -0.5 * ncoil * thickness
	coil = Wire([	vec3(-r*sin(t), r*cos(t), z0 + t/(2*pi) * thickness)
					for t in linrange(0, 2*pi * ncoil, step) ]) .qualify('spring')
				
	# create hooks
	c = thickness * sign(hook)
	if abs(c) > abs(hook):
		hook = c
	top = Wire([
			vec3(arm, 0, hook),
			vec3(arm, 0, c),
			vec3(arm-abs(c), 0, 0),
			vec3(0),
			]) .transform(coil[0])
	bot = (top 
			.flip() 
			.transform(scaledir(Z,-1)*scaledir(X,-1)) 
			.transform(angleAxis(angle, Z))
			)
	
	# create path
	path = top + coil + bot
	path.mergeclose()
	
	if not solid:
		return path
	
	return Solid(
			part=tube(
				flatsurface(Circle(
					(path[0], normalize(path[0]-path[1])),
					thickness/2,
					resolution=('div',6),
					)),
				path,
				),
			axis=Axis(O,Z, interval=(-length/2, length/2)),
			top=path[0],
			bottom=path[-1],
			)
		
		
		
# ----------------------- bearing stuff --------------------------

@cachefunc
def bearing(dint, dext=None, h=None, 
			circulating='ball', 
			contact=0, 
			hint=None, hext=None, 
			sealing=False, 
			detail=False) -> Solid:
	''' 
		Circulating bearings rely on rolling elements to avoid friction and widen the part life.
		Its friction depends on the rotation speed but not on the current load.
		
		See bearing specs at https://koyo.jtekt.co.jp/en/support/bearing-knowledge/
		
		Parameters:

			dint:	interior bore diameter
			dext:  exterior ring diameter
			h:     total height of the bearing
			
			hint:  height of the interior ring. Only for angled roller bearings
			hext:  height of the exterior ring. Only for angled roller bearings
		
			circulating:
			
				The type of circulating element in the bearing
				
				- ball
				- roller
				
			contact:
			
				Contact angle (aka pressure angle).
				It decided what directions of force the bearing can sustain.
				
					- `0` for radial bearing
					- `pi/2`  for thrust (axial) bearings
					- `0 < contact < pi/2` for conic bearings
					
			sealing:  True if the bearing has a sealing side. Only for balls bearings with `contact = 0`
					
			detail:   If True, the returned model will have the circulating elements and cage, if False the returned element is just a bounding representation
	'''
	if isinstance(dint, str):
		dint, dext, h = bearing_spec(dint)
	else:
		if not dext:	dext = 2*dint
		if not h:		h = 0.5*dint
		
	assert 0 < h <= dint < dext
	assert 0 <= contact

	if circulating == 'roller':
		if sealing:	
			raise ValueError('sealing must be None for roller bearings')
		return bearing_roller(dint, dext, h, contact, hint, hext, detail)
	
	elif circulating == 'ball':
		if hint or hext:
			raise ValueError('hint and hext must be None for ball bearings')
		if abs(contact) < NUMPREC:
			return bearing_ball(dint, dext, h, sealing, detail)
		if abs(contact - 0.5*pi) < NUMPREC:
			if sealing:
				raise ValueError('sealing must be none for thrust bearings')
			return bearing_thrust(dint, dext, h, detail)
		else:
			raise ValueError('ball bearings must not be angled')
			
	else:
		raise ValueError('unknown circulating element {}'.format(repr(circulating)))
	
	
def bearing_spec(code):
	# iso code
	if ' ' in code:
		indev
	# simple code
	elif re.match(r'[\d\.]+x[\d\.]+x[\d\.]+', code):
		return tuple(float(d) for d in re.split('x'))
		
		
def bearing_ball(dint, dext=None, h=None, sealing=False, detail=False) -> Solid:
	# convenient variables
	rint = dint/2
	rext = dext/2
	c = 0.05*h
	w = 0.5*h
	e = 0.15*(dext-dint)

	# outer rings profiles
	axis = Axis(O,Z, interval=(0,h))
	interior = Wire([
		vec3(rint+e, 0,	w), 
		vec3(rint, 0, w),
		vec3(rint,	0,	-w),
		vec3(rint+e, 0,	-w), 
		]) .segmented() .flip()
	filet(interior, [1, 2], radius=c, resolution=('div',1))

	exterior = Wire([
		vec3(rext-e,	0, -w),
		vec3(rext, 0, -w),
		vec3(rext, 0, w),
		vec3(rext-e,	0, w),
		]) .segmented() .flip()
	filet(exterior, [1,2], radius=c, resolution=('div',1))

	if detail and not sealing:
		rb = (dint + dext)/4	# balls path radius
		rr = 0.75*(dext - dint)/4 # balls radius

		hr = sqrt(rr**2 - (rb-rint-e)**2)	# half balls inprint width in the rings
		interior += wire(ArcCentered((rb*X,-Y), vec3(rint+e, 0, hr), vec3(rint+e, 0, -hr)))
		exterior += wire(ArcCentered((rb*X,-Y), vec3(rext-e, 0, -hr), vec3(rext-e, 0, hr)))
		interior.close()
		exterior.close()
		part = revolution(web([exterior, interior]), axis) .option(color=settings.colors['bearing'])

		nb = int(0.7 * pi*rb/rr)	# number of balls that can fit in
		balls = repeat(icosphere(rb*X, rr), nb, angleAxis(radians(360)/nb, Z)) .option(color=vec3(0,0.1,0.2))
		
		# balls cage (simplified version)
		cage_profile = Wire([vec3(
						rb*cos(t), 
						rb*sin(t), 
						(rr+0.5*c) * (1 - (0.5-0.5*cos(nb*t))**2) + c)
					for t in linrange(0, 2*pi, div=nb*20-1)])
		cage_profile.mergeclose()

		surf = extrusion(
				cage_profile, 
				mat3(
					(rb-rr*0.6)/rb, 
					(rb-rr*0.6)/rb, 
					1), 
				alignment=0.5)

		cage = thicken(
				surf + surf.transform(mat3(1,1,-1)) .flip(), 
				c) .option(color=settings.colors['bearing_cage'])
				
		# asemble
		return Solid(part=part, cage=cage, balls=balls, axis=axis)

	else:
		# assemble and close rings profiles
		interior = (
			Wire([
				exterior[-1], 
				exterior[-1]+c*Z, 
				interior[0]+c*Z, 
				interior[0]]) .segmented()
			+ interior
			+ Wire([
				interior[-1], 
				interior[-1]-c*Z, 
				exterior[0]-c*Z, 
				exterior[0]]) .segmented()
			)
		return Solid(
				part=revolution(web([exterior, interior]), axis)
						.option(color=settings.colors['bearing']), 
				axis=axis)


def bearing_roller(dint, dext=None, h=None, contact=0, hint=None, hext=None, detail=False) -> Solid:
	# interior and exterior heights (automatically deduced from total height if not specified)
	if not hint:	hint = h*cos(contact)
	if not hext:	hext = h*cos(contact) * 0.8 if contact else h
	
	assert 0 < hint <= h
	assert 0 < hext <= h

	# convenient variables
	rint = dint/2
	rext = dext/2
	c = 0.05*h
	w = 0.5*h
	e = 0.08*(dext-dint)
	axis = Axis(O,Z, interval=(0,h))
	
	# cones definition points
	if contact:
		# axis on a cone using the given contact angle
		cr = vec3(mix(rint, rext, 0.5), 0, 0)
		ct = -cr.x / tan(contact) *Z
		angled = Axis(ct, vec3(sin(contact), 0, cos(contact)))

		p1 = vec3(rext-e, 0, -w+hext)
		p2 = vec3(rint+e, 0, w-hint+e)
		a1 = Axis(p1, normalize(p1-ct))
		a2 = Axis(p2, normalize(p2-ct))

		p3 = angled[0] - project(reflect(p2-angled[0], angled[1]), a1[1])

		p4 = angled[0] - reflect(p3-angled[0], angled[1])
		p5 = angled[0] - reflect(p1-angled[0], angled[1])
	else:
		# cylindrig bearing: infinite cone
		p1 = vec3(rext-e, 0, -w+hext-e)
		p3 = vec3(rext-e, 0, -w+e)
		p4 = vec3(rint+e, 0, w-hint+e)
		p5 = vec3(rint+e, 0, -w+hext-e)
		angled = Axis(0.5*(rint+rext) * X, Z)

	# ring profiles
	interior = Wire([
		p5+e*X,
		vec3(p5[0]+e, 0, w),
		vec3(rint, 0, w),
		vec3(rint,	0,	w-hint),
		vec3(p4[0]+e, 0, w-hint),
		p4+e*X,
		]) .segmented()
	exterior = Wire([
		p3,
		vec3(p3[0], 0, -w),
		vec3(rext, 0, -w),
		vec3(rext, 0, -w+hext),
		vec3(rext-e,	0, -w+hext),
		]) .segmented()
	filet(interior, [2,3], radius=c, resolution=('div',1))
	filet(exterior, [2,3], radius=c, resolution=('div',1))
	
	# create interior details
	if detail:
		# complete rings with their interiors
		interior += Wire([
			p4,
			p5,
			p5+e*X,
			]) .segmented()
		exterior += Wire([p3])

		part = revolution(web([
					exterior, 
					interior,
					]).flip(), axis) .option(color=settings.colors['bearing'])
	
		# create conic rollers
		roller = revolution(Segment(mix(p1,p3,0.05), mix(p3,p1,0.05)), angled)
		roller.check()
		for hole in roller.outlines().islands():
			roller += flatsurface(wire(hole))
			roller.check()

		# number of rollers that can fit in
		nb = int(pi*(rint+rext) / (2.5*distance_pa(p1,angled)))
		rollers = repeat(roller, nb, rotatearound(2*pi/nb, axis)) 
		rollers.option(color=settings.colors['circulating'])

		# roller cage
		p6 = mix(p4,p3,0.6) - e*angled[1]
		cage_profile = wire([
			p6 - 1.5*e*X,
			p6,
			p6 + (distance(p1,p3)+1.8*e) * angled[1],
			])
		filet(cage_profile, [1], radius=c)
		cage = revolution(cage_profile, axis)
		cage = pierce(cage, inflate(rollers, 0.5*c), False)
		cage = thicken(cage, c) .option(color=settings.colors['bearing_cage'])
		
		# assemble
		return Solid(part=part, cage=cage, rollers=rollers, axis=axis)
	
	# simply create a bounding representation
	else:
		# assemble and close rings profiles
		return Solid(
				part=revolution(
					(exterior + interior) .close() .flip(),
					axis,
					) .option(color=settings.colors['bearing']), 
				axis=axis)
		

def bearing_thrust(dint, dext, h, detail=False) -> Solid:
	# convenient variables
	rint = dint/2
	rext = dext/2
	c = 0.05*h	# corner radius
	w = 0.5*h	# half height
	e = 0.12*(dext-dint) # outer rings thickness

	# rings outlines
	axis = Axis(O,Z, interval=(0,h))
	top = Wire([
		vec3(rext, 0, w-e), 
		vec3(rext, 0, w),
		vec3(rint, 0, w),
		vec3(rint, 0, w-e), 
		]) .segmented() .flip()
	filet(top, [1, 2], radius=c, resolution=('div',1))

	bot = Wire([
		vec3(rint, 0, -w+e),
		vec3(rint, 0, -w),
		vec3(rext, 0, -w),
		vec3(rext, 0, -w+e),
		]) .segmented() .flip()
	filet(bot, [1,2], radius=c, resolution=('div',1))

	if detail:
		rb = (dint + dext)/4	# balls guide radius
		rr = 0.75*h/2		# ball radius
		
		hr = sqrt(rr**2 - (w-e)**2)		# half ball inprint width in the rings
		# complete the rings with their interior
		top += wire(ArcCentered((rb*X,-Y), vec3(rb+hr, 0, w-e), vec3(rb-hr, 0, w-e)))
		bot += wire(ArcCentered((rb*X,-Y), vec3(rb-hr, 0, -w+e), vec3(rb+hr, 0, -w+e)))
		top.close()
		bot.close()
		part = revolution(web([top, bot]), axis) .option(color=settings.colors['bearing'])
		
		# number of balls to place
		nb = int(0.8 * pi*rb/rr)
		balls = repeat(icosphere(rb*X, rr), nb, angleAxis(radians(360)/nb, Z))
		balls.option(color=settings.colors['circulating'])
		
		# cage
		cage_profile = Wire([ 
			vec3(rext-c, 0, -w+e+0.1*h),
			vec3(rext-c, 0, w-e-0.1*h),
			vec3(rint+c, 0, w-e-0.1*h),
			vec3(rint+c, 0, -w+e+0.1*h),
			])
		filet(cage_profile, [1,2], radius=c, resolution=('div',1))
		
		cage_surf = revolution(cage_profile, axis)
		cage_surf = pierce(cage_surf, inflate(balls, 0.2*c), False)
		cage = thicken(cage_surf, c) .option(color=settings.colors['bearing_cage'])
		
		# assemble
		return Solid(part=part, cage=cage, balls=balls, axis=axis)
		
	else:
		# assemble and close the profiles
		top = (
			Wire([
				bot[-1], 
				bot[-1]+c*X, 
				top[0]+c*X, 
				top[0]]) .segmented()
			+ top
			+ Wire([
				top[-1], 
				top[-1]-c*X, 
				bot[0]-c*X, 
				bot[0]]) .segmented()
			)

		return Solid(
				part=revolution(web([top, bot]), axis) .option(color=settings.colors['bearing']), 
				axis=axis)

	
'''
	all existing bearing dimensions according to https://koyo.jtekt.co.jp/en/support/bearing-knowledge/6-3000.html
'''
standard_bearing_ball_straight = [
	(16, 35, 11),
	]



from .selection import *

@cachefunc
def slidebearing(dint, h=None, thickness=None, shoulder=None, opened=False) -> Solid:
	'''
		Slide bearings rely on gliding parts to ensure a good pivot. It's much cheaper than circulating bearings and much more compact. But needs lubricant and has a shorter life than circulating bearings.
		Its friction depends on the rotation speed and on the load.
	
		Parameters:
			dint:       interior diameter
			h:          exterior height (under shoulder if there is)
			thickness:  shell thickness, can be automatically determined
			shoulder:   distance from bore to shoulder tip, put 0 to disable
			opened:     enable to have a slight breach that allows a better fitting to the placement hole 
	'''
	if h is None:			h = 1.2*dint
	if thickness is None:	thickness = 0.05*dint
	rint = dint/2
	
	profile = Wire([
				vec3(rint,0,-h),
				vec3(rint,0, 0),
				])
	if shoulder:
		profile += Wire([ vec3(rint+shoulder, 0, 0) ])
		profile = profile.segmented()
		filet(profile, [1], radius=1.5*thickness, resolution=('div',2))

	axis = Axis(O,Z, interval=(-h,0))
	shape = revolution(profile, axis, 1.98*pi if opened else 2*pi)
	
	part = thicken(shape, thickness) .option(color=settings.colors['bearing'])
	line = (  select(part, vec3(-rint,0,-h), stopangle(pi/2))
			+ select(part, vec3(dint,dint,-h), stopangle(pi/2))
			)
	if not shoulder:
		line += (  select(part, vec3(-rint,0,0), stopangle(pi/2))
				 + select(part, vec3(dint,dint,0), stopangle(pi/2))
				)
	
	chamfer(part, line, width=0.5*thickness)
	return Solid(part=part, axis=axis)

	
def bolt(a: vec3, b: vec3, dscrew: float, washera=False, washerb=False, nutb=True) -> Solid:
	''' convenient function to create a screw, nut and washers assembly 
	
		Parameters:
			a:  the screw placement
			b:  the nut placement
			dscrew:  the screw `d` parameter
			washera:  if True, place a washer between the screw head at `a`
			washerb:  if True, place a washer between the nut at `b`
			butb:    if True, place a nut at `b`
	'''
	dir = normalize(b-a)
	rwasher = washer(dscrew)
	thickness = rwasher['part'].box().width.z
	rscrew = screw(dscrew, stceil(distance(a,b) + 1.2*dscrew, precision=0.2))
	#rscrew['annotations'] = [
				#note_distance(O, -stceil(distance(a,b) + 1.2*dscrew)*Z, offset=2*dscrew*X),
				#note_radius(rscrew['part'].group(0)),
				#]
	result = Solid(
			screw = rscrew.place((Revolute, rscrew['axis'], Axis(a-thickness*dir*int(washera), -dir))), 
			)
	if washera:
		result['washera'] = rwasher.place((Revolute, rwasher['top'], Axis(a, dir)))
	if washerb:
		result['washerb'] = rwasher.place((Revolute, rwasher['top'], Axis(b, -dir)))
	if nutb:
		rnut = nut(dscrew)
		result['nut'] = rnut.place((Revolute, rnut['top'], Axis(b+thickness*dir*int(washerb), -dir)))
	return result

def screw_slot(axis: Axis, dscrew: float, rslot=None, hole=0., screw=0., expand=True, flat=False) -> Mesh:
	''' slot shape for a screw
		the result can then be used in a boolean operation to reserve set a screw place in an arbitrary shape
		
		Parameters:
			axis:  the screw axis placement, z toward the screw head (part exterior)
			dscrew: the screw diameter
			rslot:  the screw head slot radius
			hole:   
				- if `True`, enables a cylindric hole for screw body of length `dscrew*3`
				- if `float`, it is the screw hole length
			screw:  if non zero, this is the length of a thiner portion of hole after `hole`, the diameter is adjusted so that the screw can screw in
			expand: 
				- if `True`, enables slots sides
				- if `float`, it is the slot sides height
			flat:   if True, the slot will be conic do receive a flat head screw
	'''
	if not rslot:	rslot = 1.1*dscrew
	o = axis[0]
	x,y,z = dirbase(axis[1])

	profile = []
	if expand:
		if isinstance(expand, bool):		expand = 2*rslot
		profile.append(o + rslot*x + expand*z)
	profile.append(o + rslot*x)
	if hole or screw:
		if flat:
			profile.append(o + 0.5*dscrew*(x-z) - (rslot-dscrew)*z)
			hole = max(hole, dot(o-profile[-1], z))
		else:
			profile.append(o + 0.5*dscrew*x)
		profile.append(o + 0.5*dscrew*x - hole*z)
		profile.append(o + 0.4*dscrew*x - (hole+0.1*dscrew)*z)
		profile.append(o + 0.4*dscrew*x - (hole+screw)*z)
		profile.append(o - (hole+screw+0.4*dscrew)*z)
	else:
		profile.append(o)
	return revolution(wire(profile).segmented(), Axis(o,-z)).finish()

def bolt_slot(a: vec3, b: vec3, dscrew: float, rslot=None, hole=True, expanda=True, expandb=True) -> Mesh:
	''' bolt shape for a screw
		musch like `screw_slot()` but with two endings
		
		Parameters:
			a:  position of screw head
			b:  position of nut
			dscrew:  the screw diameter
			rslot:   the screw head slot radius
			hole:    enabled the cylindric hole between the head and nut slots
			expanda, expandb:
				- if `True`, enables slot sides on tip `a` or `b`
				- if `float`, it is the slot side height
				
		Example:
			
			>>> a, b = vec3(...), vec3(...)
			>>> bolts = bolt(a, b, 3)
			>>> part_hole = bolt_slot(a, b, 3)
	'''
	if not rslot:	rslot = 1.3*dscrew
	x,y,z = dirbase(normalize(a-b))

	def one_slot(o,x,z, expand):
		profile = []
		if expand:
			if isinstance(expand, bool):		expand = 2*rslot
			profile.append(o + rslot*x + expand*z)
		profile.append(o + rslot*x)
		if hole:
			profile.append(o + 0.5*dscrew*x)
		else:
			profile.append(o)
		return wire(profile).segmented()

	return revolution(
			one_slot(a,x,z, expanda) + one_slot(b,x,-z, expandb).flip(),
			Axis(a,-z),
			).finish()

def bearing_slot_exterior(axis: Axis, dext: float, height: float, shouldering=True, circlip=False, evade=True, expand=True):
	''' slot for a bearing exterior ring, such as returned by `bearing()`
		
		Parameters:
			axis:  bearing axis placement
			dext:  bearing `dext` parameter
			height: bearing `h` parameter
			shouldering:  generate a shoulder to support the top side of the bearing's exterior ring
			circlip: enable a slot for a circlip to support the bottom side of the bearing's exterior ring
			evade:  set expansion to evasive rather that straight
			expand:
			
				expand the slot with evasive or straight surfaces to ease itnersection with other meshes
				
				- `True` enables expansion with a default length
				- `float` set the expansion distance
	'''
	rext = dext/2
	tol = stceil(rext*1e-2)
	profile = [
		rext*X + 0.5*height*Z,
		rext*X - 0.5*height*Z,
		]
	if shouldering:
		profile.insert(0, profile[0] - 0.17*rext*X)
	if circlip:
		circlip_height = stceil(0.1*rext)
		circlip_depth = stceil(1.05*rext) - rext
		profile[-1:] = accumulate([
			rext*X - (0.5*height + tol)*Z,
			circlip_depth*X,
			-circlip_height*Z,
			-circlip_depth*X,
			-0.5*circlip_height*Z,
			])
	if expand == True:	expand = 0.4*rext
	if evade:
		if shouldering:
			profile.insert(0, profile[0] + 0.1*rext*Z)
		profile.append(profile[-1] + expand*(X-Z))
		profile.insert(0, profile[0] + expand*(X+Z))
	elif expand:
		profile.extend([
			profile[-1] - 0.05*rext*(Z-X),
			profile[-1] - 0.05*rext*(Z-X) - expand*Z,
			])
		profile.insert(0, profile[0] + expand*Z)
	return revolution(wire(profile).segmented(), Axis(O,-Z)) .transform(translate(axis[0]) * mat4(quat(Z, axis[1])))

def bearing_slot_interior(axis: Axis, dint: float, height: float, shouldering=True, circlip=False, evade=True, expand=True) -> Mesh:
	''' slot for a bearing interior ring, such as returned by `bearing()`
	
		Parameters:
			axis:  bearing axis placement
			dint:  bearing `dint` parameter
			height:  bearing `h` parameter
			shouldering:  generate a shoulder to support the top side of the bearing's interior ring
			circlip:  generate a slot for a circlip to support the bottom side of the bearing's interior ring
			evade:  set expansion to evasive rather that straight
			expand: 
			
				expand the slot with evasive or straight surfaces to ease itnersection with other meshes
				
				- `True` enables expansion with a default length
				- `float` set the expansion distance
	'''
	rint = dint/2
	tol = stceil(rint*2e-2)
	profile = [
		rint*X + 0.5*height*Z,
		rint*X - 0.5*height*Z,
		]
	if shouldering:
		profile.insert(0, profile[0] + 0.36*rint*X)
	if circlip:
		circlip_height = stceil(0.2*rint)
		circlip_depth = rint - stceil(0.8*rint)
		profile[-1:] = accumulate([
			rint*X - (0.5*height + tol)*Z,
			-circlip_depth*X,
			-circlip_height*Z,
			circlip_depth*X,
			-0.5*circlip_height*Z,
			])
	if expand == True:	expand = 0.8*rint
	if evade:
		if shouldering:
			profile.insert(0, profile[0] + 0.1*rint*Z)
		profile.append(profile[-1] - expand*(X+Z))
		profile.insert(0, profile[0] - expand*(X-Z))
	elif expand:
		profile.extend([
			profile[-1] - 0.1*rint*(Z+X),
			profile[-1] - 0.1*rint*(Z+X) - expand*Z,
			])
		profile.insert(0, profile[0] + expand*Z)
	return revolution(wire(profile).segmented(), Axis(O,Z)) .transform(translate(axis[0]) * mat4(quat(Z, axis[1])))


def circular_screwing(axis, radius, height, dscrew, diameters:int=1, div:int=8, hold:float=0) -> '(Mesh, list)':
	''' holes and bolts to screw a circular perimeter
	
		Parameters:
			axis:  placement of circle around which the screws are placed
			radius: radius of the perimeter on which to put the screws
			height: bolts length
			dscrew: diameter of the biggest screws
			diameters: each biggest screw is followed by this number-1 of additional smaller diameters
			div: number of biggest screws distributed uniformly on the perimeter
			hold: 
				if non zero, holding screws of the given length are added to help the assembly.
				if True, the holding screws length is automatically determined.
				
		Return: a tuple (holes:Mesh, bolts:list) where the bolts is a list of `Solid`
	'''
	holes = Mesh()
	bolts = []
	x,y,z = dirbase(axis.direction)
	a = axis.origin + radius*x
	b = axis.origin + radius*x + height*z
	gap = 0.1*dscrew*z
	enlarge = 1.05
	bolts.append(bolt(a, b, dscrew))
	for i in range(diameters):
		holes += cylinder(a-gap, b+gap, enlarge*stfloor(0.8**i * dscrew)/2) .transform(rotatearound(i*1.4*dscrew/radius, axis))
	holes = repeataround(holes, 8, axis).flip()
	if hold is True:
		hole = 1.5*dscrew
	if hold:
		angle = 1.7*dscrew/radius
		holes += repeataround(screw_slot(Axis(a+gap,-z), dscrew, 
					screw=height-hold, 
					hole=hold, 
					flat=True), 2) .transform(rotatearound(-angle, axis))
		bolts.append(screw(dscrew, height, head='flat')
			.place((Revolute, Axis(O,Z), Axis(a,-z)))
			.transform(rotatearound(-angle, axis)))
		bolts.append(screw(dscrew, height, head='flat')
			.place((Revolute, Axis(O,Z), Axis(a,-z)))
			.transform(rotatearound(pi-angle, axis)))
	return holes, bolts
	

def grooves_profile(radius, repetitions:int=16, alignment=0.5, angle=radians(40)) -> Wire:
	''' Coupling grooves profile 
		
		- The size of grooves depend on the pressure angle and the number of grooves
		
		Parameters:
		
			radius: the radius around which the grooves are placed
			repetitions: the number of grooves to put on the circle perimeter
			alignemnt: fraction of the grooves height
				- exterior grooves usually have alignment=1
				- interior grooves usually have alignemnt=0
			angle: pressure angle of the grooves
	'''
	h = radius/repetitions / tan(angle)
	def profile(t, min=-1, max=1):
		return (radius+(0.5-alignment)*h+h*sin(repetitions*t)) * vec3(cos(t), sin(t), 0)
	return wire([ profile(t, -1, 0.7)
		for t in linrange(0, 2*pi, div=12*repetitions, end=False) ]).close()

def grooves(radius, height, repetitions:int=16, alignment=0.5, angle=radians(40)) -> Mesh:
	''' Coupling grooves surface
		
		- The bottom and top bevels direction depend on the alignment
		- The size of grooves depend on the pressure angle and the number of grooves
		
		Parameters:
		
			radius: the radius around which the grooves are placed
			height: the length of the grooves, including transition bevels
			repetitions: the number of grooves to put on the circle perimeter
			alignemnt: fraction of the grooves height
				- exterior grooves usually have alignment=1
				- interior grooves usually have alignemnt=0
			angle: pressure angle of the grooves
	'''
	h = radius/repetitions / tan(angle)
	if alignment > 0.5:    s = 1+h/radius
	else:                  s = 1-h/radius
	return extrans(grooves_profile(radius, repetitions, alignment, angle), [
				scale(vec3(s)),
				translate(h*Z),
				translate((height-h)*Z),
				translate(height*Z) * scale(vec3(s)),
				])
	

from .kinematic import Chain, Kinematic
from .joints import *
	
def scara(backarm:float, forearm:float) -> Kinematic:
	'''
	kinematic of a classical *scara* robot arm
	'''
	return Chain([
		Revolute(('base',1), Axis(O,Z)),
		Revolute((1,2), Axis(backarm*X,Z), Axis(O,Z)),
		Revolute((2,3), Axis(forearm*X,Z), Axis(O,Z)),
		Prismatic((3,'tool'), Axis(-forearm*0.5*Z,Z)),
		])

def serial6(backarm:float, forearm:float) -> Kinematic:
	''' 
	kinematic of a classical serial robot arm with 6 degrees of freedom
	
	Parameters:
		backarm:  backarm length
		forearm:  forearm length
	'''
	return Chain([
		Revolute(('base',1), Axis(-0.3*backarm*Z,Z)),  # shoulder 1
		Revolute((1,2), Axis(O,Y)),  # shoulder 2
		Revolute((2,3), Axis(backarm*Z,Y), Axis(O,Y)), # elbow
		Revolute((3,4), Axis(0.5*forearm*Z,Z)),  # wrist 1
		Revolute((4,5), Axis(forearm*Z,Y), Axis(O,Y)), # wrist 2
		Revolute((5,'tool'), Axis(0.3*forearm*Z,Z)), # wrist 3
		])

def serial7(backarm:float, forearm:float) -> Kinematic:
	''' 
	kinematic of a classical serial robot arm with 7 degrees of freedom
	
	Parameters:
		backarm:  backarm length
		forearm:  forearm length
	'''
	return Chain([
		Revolute(('base',1), Axis(-0.3*backarm*Z,Z)),  # shoulder 1
		Revolute((1,2), Axis(O,Y)),  # shoulder 2
		Revolute((2,3), Axis(0.5*backarm*Z,Z)),  # 7th degree
		Revolute((3,4), Axis(backarm*Z,Y), Axis(O,Y)), # elbow
		Revolute((4,5), Axis(0.5*forearm*Z,Z)),  # wrist 1
		Revolute((5,6), Axis(forearm*Z,Y), Axis(O,Y)), # wrist 2
		Revolute((6,'tool'), Axis(0.3*forearm*Z,Z)), # wrist 3
		])

def delta3(base:float, tool:float, backarm:float, forearm:float, forearm_width:float=None) -> Kinematic:
	'''
	kinematic of a classical delta robot with 3 degrees of freedom
	
	Parameters:
		base:   the radius of motors placements
		tool:   the radius of the tool's ball joints placements
		backarm:  backarm length
		forearm:  forearm length
		forearm_width:  gap between the 2 forarm segments
	'''
	if forearm_width is None:	forearm_width = 0.1*forearm
	from . import generation as gt
	arms = gt.regon(Axis(O,Z), 1, 3).points
	motors = []
	joints = []
	for i, p in enumerate(arms):
		motors.append(Revolute(('base', f'backarm.{i}'), Axis(p,cross(Z, p*base)), Axis(O,X)))
		joints.append(Ball((f'backarm.{i}', f'forearm.{i}.0'), backarm*Z + forearm_width*X, O))
		joints.append(Ball((f'backarm.{i}', f'forearm.{i}.1'), backarm*Z - forearm_width*X, O))
		joints.append(Ball(('tool', f'forearm.{i}.0'), p*tool + forearm_width*cross(Z,p), forearm*Z))
		joints.append(Ball(('tool', f'forearm.{i}.1'), p*tool - forearm_width*cross(Z,p), forearm*Z))
		
		# joints.append(Ball((f'backarm.{i}', f'forearm.{i}.0'), backarm*Z + forearm_width*X, O))
		# joints.append(Ball((f'backarm.{i}', f'forearm.{i}.1'), backarm*Z - forearm_width*X, O))
		# joints.append(Ball(('tool', f'forearm.{i}.0'), p*tool + forearm_width*cross(Z,p), forearm*Z))
		# joints.append(Ball(('tool', f'forearm.{i}.1'), p*tool - forearm_width*cross(Z,p), forearm*Z))
	return Kinematic(joints, inputs=motors, outputs=['tool'], ground='base')
	# return Kinematic(motors+joints, ground='base')

# def delta4(backarm, forearm) -> Kinematic:
# 	indev
# def delta6(backarm, forearm) -> Kinematic:
# 	indev
	
# def stewart(base, tool, shaft, gap=None):
# 	if gap is None:	gap = base*0.1
# 	shafts = gt.regon(Axis(O,Z), 1, 3).points
# 	motors = []
# 	joints = []
# 	for i in range(len(shafts)):
# 		motors.append(Cylindrical((f'shaft.{i}.base', f'shaft.{i}.tool'), Axis(0.5*arm*Z, Z)))
# 		joints.append(UniversalJoint(('base', f'shaft.{i}.base'), Axis(base*directions[i] - gap*cross(Z,directions[i]), directions[i]), Axis(O,X)))
# 		joints.append(UniversalJoint(('base', f'shaft.{i+1}.base'), Axis(base*directions[i] + gap*cross(Z,directions[i]), directions[i]), Axis(O,X)))
	
'''
* profilés
	+ ipn
	+ (carrés a rails)
	+ U
	+ L
* nut
	+ carrés
	+ hex
	+ auto-stopable
	+ fermés au bout
	+ support
* screw
	+ forme tete (UH, BH, VH, ...)
	+ forme empreinte (torx, hex, crux, line, ...)
* rondelle
	+ simple
	+ stobable
	+ ressort
	+ entretoise
* roulement
	+ bille, rouleau
	+ droit, oblique, radial
	+ etancheité
* coussinet
	+ droit
	+ a epaule
	+ fendu
* anneaux elastiques
	+ circlip interieur/exterieur
	+ ressort
	
* helice
	+ rotor de ventilateur ...
	
* vis/ecrou
	+ a billes
	+ a profil trapezoidal
* accouplements
	+ ressort
	+ encastrement: prisme, cannelures, clavete, meplat
* ressorts
	+ lineaire
	+ extension
	+ conique
	+ pli
	+ spiral
	+ concentrique
	+ lame
	
* tuyau
	+ embouchure
	+ droit/coude
	+ le long d'un chemin ?
	
pour tout:
	- liste des tailles standard
'''
	



# This file is part of pymadcad,  distributed under license LGPL v3
''' This module provides the chamfer and filet functions, and the associated tools.

	`edgecut`, `chamfer`, and `filet` are built using the same cutting algorithm. It cuts the mesh faces by propagation from the given edges. The cutting planes are determined by an offset vector from the original primitive (point or edge). 
	
	Most of the time you don't need to set the offset yourself. It can be automatically calculated by several methods, depending on the shape you want to get. Those methods are called `cutters` and are executed using `planeoffsets`.
'''

from .mathutils import *
from .mesh import *
from .hashing import *
from . import generation as gt
from . import hashing
from . import text
from . import settings
from .triangulation import triangulation_outline
from .blending import blenditer, match_length

from numbers import Number
from functools import singledispatch, partial

__all__ = [	'chamfer', 'filet', 'edgecut',
			'mesh_cut', 'web_cut', 'planeoffsets',
			'tangentend', 'tangentcorner', 'tangentjunction',
			'cutter_width', 'cutter_distance', 'cutter_depth', 'cutter_radius',
			]
			
			
# ---- common cut methods ----

@singledispatch
def edgecut(mesh, indices, **cutter):
	''' Cut a Mesh/Web/Wire around the given edges/points, using the given cutter '''
	raise TypeError('wrong argument type: {}'.format(type(mesh)))

@singledispatch
def chamfer(mesh, indices, **cutter):
	''' Cut the given edges or points on the mesh and replace them by a chamfer: a flat surface filling the hole
	
	Example
	
		>>> cube = brick(width=5)
		>>> chamfer(cube, cube.frontiers(), width=1)
	'''
	raise TypeError('wrong argument type: {}'.format(type(mesh)))
	
@singledispatch
def filet(mesh, indices, **cutter):
	''' Cut the given edges or points on the mesh and replace them by a filet: a round surface tangenting and filling the hole
	
	Example
	
		>>> cube = brick(width=5)
		>>> filet(cube, cube.frontiers(), width=1)
	'''
	raise TypeError('wrong argument type: {}'.format(type(mesh)))




# ---- cut methods -----

def cutter_width(width, fn1, fn2):
	''' Plane offset for a cut based on the width of the filet '''
	n = normalize(fn1+fn2)
	s = dot(fn1,n)
	return -width/2 * sqrt(1/s**2 - 1) * n

def cutter_distance(depth, fn1, fn2):
	''' Plane offset for a cut based on the distance to the cutted edge '''
	n = normalize(fn1 +fn2)
	cos_b = dot(fn1, n)
	cos_a = sqrt(1-cos_b**2)
	return -depth * n * cos_a

def cutter_depth(dist, fn1, fn2):
	"""Plane offset for a cut based on the distance along the side faces"""
	return -dist * cross(normalize(cross(fn1,fn2)), fn1-fn2)

def cutter_radius(depth, fn1, fn2):
	''' Plane offset for a cut based on the angle between faces '''
	n = normalize(fn1 + fn2)
	s = dot(fn1,n)
	return -depth * (1/s - s) * n




# ----- mesh operations ------

def planeoffsets(mesh, edges, **cutter):
	''' Compute the offsets for cutting planes using the given method 
		cutter is a tuple or a function
		
			- function(fn1,fn2) -> offset 		
				fn1, fn2 are the adjacents face normals
				offset is the distance from segment to plane times the normal to the plane
				
			- ('method', distance) 				
				the method is the string name of the method (a function named 'cutter_'+method existing in this module)
				distance depends on the method and is the numeric parameter of the method
	'''
	cutter = interpretcutter(**cutter)
	# get adjacent faces' normals to lines
	offsets = {e: [None,None]  for e in edges}
	for f in mesh.faces:
		for e in ((f[0],f[1]), (f[1],f[2]), (f[2],f[0])):
			if e in offsets:					offsets[e][0] = mesh.facenormal(f)
			elif (e[1],e[0]) in offsets:		offsets[(e[1],e[0])][1] = mesh.facenormal(f)
	
	# compute offsets (displacement on depth)
	for e, adjacents in offsets.items():
		offset = cutter(*adjacents)
		if dot(cross(*adjacents), mesh.points[e[0]]-mesh.points[e[1]]) > 0:
			offset = -offset
		offsets[e] = offset
	
	return offsets

def interpretcutter(**cutter):
	if len(cutter) != 1:
		raise TypeError("cutting method must be exactly one keyword argument")
	if callable(func := cutter.get('cutter')):
		return func
	else:
		name, dim = next(iter(cutter.items()))
		func = globals().get('cutter_'+name)
		if not callable(func):
			raise TypeError("no cutting method named {}".format(repr(name)))
		return partial(func, dim)

		
def mesh_cut(mesh, start, cutplane, stops, conn, prec, removal, cutghost=True):
	''' Propagation cut for an edge 
		
		:start:		the edge or point to start propagation from
		:cutplane:	the plane cutting the faces. Its normal must be oriented toward the propagation area.
		:stops:     the planes stopping the propagation. Their normal must be oriented toward the propagation area.
		:removal:   the set in wich the function will put the indices of faces inside
		:cutghost:  whether the function should propagate on faces already marked for removal (previously or during the propagation)
	'''
	pts = mesh.points
	ptmgr = hashing.PointSet(prec, manage=pts)
	stops = list(filter(lambda e:e, stops))
	# find the intersections axis between the planes
	axis = [intersection_plane_plane(cutplane, stop)	for stop in stops]
	
	# prepare propagation
	if isinstance(start, tuple):
		front = [start, (start[1],start[0])]
	elif isinstance(start, Number):
		front = [e	for e in conn if start in e]
	else:
		raise TypeError('wrong start: {}, cut can only start from a point or an edge'.format(start))
	intersections = set()
	seen = set()
	while front:
		# propagation affairs
		frontedge = front.pop()
		
		if frontedge not in conn:	continue
		fi = conn[frontedge]
		if fi in seen:	continue
		f = mesh.faces[fi]
		
		# do not process empty faces (its a topology singularity), but propagate
		if faceheight(mesh, fi) <= prec:
			seen.add(fi)
			for j in range(3):
				front.append((f[j],f[j-1]))
			continue
		
		# find the intersection of the triangle with the common axis to the cutplane and the stop plane
		fpts = mesh.facepoints(fi)
		p = None
		for a in axis:
			# if the intersection is on a triangle's edge, p is None
			p = intersection_axis_face(a, fpts, prec)
			if p:	break
		if p:
			# mark cutplane change
			# empty faces can be generated, but they are necessary to keep the connectivity working
			pi = ptmgr.add(p)
			l = len(mesh.faces)
			mesh.faces[fi] = (f[0], f[1], pi)
			mesh.faces.append((f[1], f[2], pi))
			mesh.faces.append((f[2], f[0], pi))
			mesh.tracks.append(mesh.tracks[fi])
			mesh.tracks.append(mesh.tracks[fi])
			registerface(mesh, conn, fi)
			registerface(mesh, conn, l)
			registerface(mesh, conn, l+1)
			front.append(frontedge)
			continue
		
		# mark this face as processed
		seen.add(fi)
		
		# point side for propagation
		# True if the point is over or on the cutplane
		goodside = [False]*3
		for j,pi in enumerate(f):
			goodside[j] = 	dot(pts[pi]-cutplane[0], cutplane[1]) >= -prec
		
		# abort conditions
		if not any(goodside):
			continue
		outside = any( all( dot(pts[pi]-stop[0], stop[1]) <= -prec
							for pi in f)
						for stop in stops)
		if outside:
			continue
		
		# True if the point is over or on for all stop planes
		goodx = [False]*3
		for j,pi in enumerate(f):
			goodx[j] = all(dot(pts[pi]-stop[0], stop[1]) >= -prec  	for stop in stops)
			
		
		# intersections of triangle's edges with the plane
		cut = [None]*3
		for j in range(3):
			cut[j] = intersection_edge_plane((pts[f[j]], pts[f[j-2]]), cutplane, prec)
			
		
		# don't intersect if the 2 points are at the corner of the triangle
		for j in range(3):
			if cut[j-1] and cut[j] and distance(cut[j-1], cut[j]) <= prec:	
				cut[j] = None
	
		if all(goodside) and any(goodx):
			removal.add(fi)
		
		if fi not in removal or cutghost:	# NOTE: this is a ugly trick to make the current edgecut implementation work with the same function for corners and edges
			# cut the face
			for j in range(3):
				if cut[j] and cut[j-1]:
					# cut only if the intersection segment is in the delimited area
					if any(	dot(cut[j-1]-stop[0], stop[1]) <= prec 
						and	dot(cut[j]  -stop[0], stop[1]) <= prec
							for stop in stops):
						continue
					
					# cut the face (create face even for non kept side, necessary for propagation)
					p1 = ptmgr.add(cut[j])
					p2 = ptmgr.add(cut[j-1])
					
					l = len(mesh.faces)
					mesh.faces[fi] = (p1, f[j-2], f[j-1])
					mesh.faces.append((p1, f[j-1], p2))
					mesh.faces.append((p1, p2, f[j-0]))
					mesh.tracks.append(mesh.tracks[fi])
					mesh.tracks.append(mesh.tracks[fi])
					registerface(mesh, conn, fi)
					registerface(mesh, conn, l)
					registerface(mesh, conn, l+1)
					seen.update((fi, l, l+1))
					# mark to remove the faces outside
					if dot(pts[f[j]]-cutplane[0], cutplane[1]) < 0:
						removal.add(fi)
						removal.add(l)
						intersections.add((p2,p1))
					else:
						removal.add(l+1)
						removal.discard(fi)
						intersections.add((p1,p2))
					break
		# propagate
		for j in range(3):
			front.append((f[j],f[j-1]))
	
	return intersections
			
	
def propagate(mesh, edges, conn, prec):
	front = list(edges)
	edges = set(edges)
	seen = set()
	count = 0
	while front:
		edge = front.pop()
		if edge not in conn:	continue
		fi = conn[edge]
		if fi in seen:	continue
		seen.add(fi)
		count += 1
		if faceheight(mesh, fi) <= prec:	continue
		
		a,b,c = arrangeface(mesh.faces[fi], edge[0])
		if c in edge:	continue
		if (b,c) not in edges:		front.append((c,b))
		if (c,a) not in edges:		front.append((a,c))
	return seen

	
@edgecut.register(Mesh)
def mesh_edgecut(mesh, edges, conn=None, prec=None, removal=None, **cutter):
	''' General purpose edge cutting function.
		cut the given edges and the crossing corners, resolving the interference issues.
		
		cutter can either be a argument for planeoffsets or more directly a dict `{edgekey: offset vector}` for each passed edge.
		edges must be unoriented.
	'''
	# perpare working objects
	if conn is None:	conn = connef(mesh.faces)
	if prec is None:	prec = mesh.precision()
	if removal is None:	
		removal = set()
		final = True
	else:
		final = False
	if isinstance(edges, Web):	edges = edges.edges
	edges = [edgekey(*e)  for e in edges]
	offsets = planeoffsets(mesh, edges, **cutter)
	
	normals = mesh.vertexnormals()
	juncconn = connpp(edges)
	pts = mesh.points
	
	debmesh = mesh
	separators = {}	# plans separateurs pour chaque arete
	for i, prox in juncconn.items():
		if len(prox) != 2:	continue
		o0 = offsets[edgekey(i,prox[0])]
		o1 = offsets[edgekey(i,prox[1])]
		axis = intersection_plane_plane((pts[prox[0]]+o0, o0), (pts[prox[1]]+o1, o1), pts[i])	# fail when the two planes have the same orientation
		if axis:
			n = normalize(cross(pts[i]-axis[0], axis[1]))
			if dot(n, pts[prox[0]]-pts[i]) < 0:		n = -n
			separators[(i,prox[0])] = (axis[0], n)
			separators[(i,prox[1])] = (axis[0],-n)
			
			## display cut planes in the mesh (debug)
			#grp = len(debmesh.groups)
			#debmesh.groups.append(None)
			#m = len(debmesh.points)
			#debmesh.points.append(pts[i])
			#debmesh.points.append(axis[0]+axis[1])
			#debmesh.points.append(axis[0]-axis[1])
			#debmesh.faces.append((m,m+1,m+2))
			#debmesh.tracks.append(grp)
			
		else:
			separators[(i,prox[0])] = None
			separators[(i,prox[1])] = None
	# fix the missing separators in case of straight lines
	for e,sep in separators.items():
		i,j = e
		while not sep:
			prox = juncconn[i]
			if len(prox) != 2:	raise MeshError('unable to compute all separators')
			j,i = i, prox[0] if prox[0] != j else prox[1]
			separators[e] = sep = separators[(i,j)]
	
	corners = {}	# offset pour chaque coin
	# pour chaque jonction
	for junc, prox in juncconn.items():
		if len(prox) < 3:	continue
		# calculer le plan de distance max demandé par les aretes adjacentes
		offset = offsets.get(junc, 0.)
		# pour chaque arete adjacente (dans chaque sens)
		for b in prox:
			for c in prox:
				a = junc
				if b == c or b == a or c == a:	continue
				# calculer la distance
				ab = normalize(pts[b]-pts[a])
				ac = normalize(pts[c]-pts[a])
				n = cross(ab, ac)
				nb = cross(n, ab)
				nc = cross(ac, n)
				ob = offsets[edgekey(a,b)]
				oc = offsets[edgekey(a,c)]
				ib = dot(ob,ob) / dot(ob, nb)
				ic = dot(oc,oc) / dot(oc, nc)
				s = dot(nb,ac)
				if not s:	continue
				abl = (ib*dot(ab,ac) + ic) / s
				offset = dot(abl*ab + ib*nb, normals[a])
		# assignation du plan de coupe
		corners[junc] = offset*normals[junc]
		plane = (pts[junc] + corners[junc], -normals[junc])
		for p in prox:
			separators[(junc,p)] = plane
	
	outlines = {}
	# couper les aretes
	for edge, offset in offsets.items():
		outlines[edge] = mesh_cut(mesh, edge, 
								(pts[edge[0]]+offset, -normalize(offset)), 
								(separators.get(edge), separators.get((edge[1], edge[0]))), 
								conn, prec, removal, True)
	# couper les sommets
	for corner,offset in corners.items():
		outlines[corner] = mesh_cut(mesh, corner, 
								(pts[corner]+offset, -normalize(offset)), 
								(),
								conn, prec, removal, False)
	if final:
		frontier = []
		for edges in outlines.values():
			frontier.extend(edges)
		finalize(mesh, outlines, propagate(mesh, frontier, conn, prec), prec)
		#finalize(mesh, outlines, removal, prec)	# NOTE this is faster but also leads to not removing all faces
	return outlines

def finalize(mesh, outlines, removal, prec):
	''' Function finalizing the mesh for edgecut
		removing faces and simplifying outlines
	'''
	# simplify cuts
	merges = {}
	for e,cuts in outlines.items():
		reindex = line_simplification(Web(mesh.points, cuts), prec)
		lines = []
		for a,b in cuts:
			c,d = reindex.get(a,a), reindex.get(b,b)
			if c != d:	lines.append((c,d))
		outlines[e] = lines
		merges.update(reindex)
	# apply merges, but keeping the empty faces, to keep the removal list valid
	for i,f in enumerate(mesh.faces):
		mesh.faces[i] = (
			merges.get(f[0],f[0]),
			merges.get(f[1],f[1]),
			merges.get(f[2],f[2]),
			)
	
	# delete inside faces and empty ones
	removefaces(mesh, lambda fi: fi in removal or faceheight(mesh,fi) <= prec)

def arrangeface(f, p):
	if   p == f[1]:	return f[1],f[2],f[0]
	elif p == f[2]:	return f[2],f[0],f[1]
	else:			return f

def registerface(mesh, conn, fi):
	f = mesh.faces[fi]
	if f[0] == f[1] or f[1] == f[2] or f[2] == f[0]:
		return
	for e in ((f[0],f[1]), (f[1],f[2]), (f[2],f[0])):
		conn[e] = fi
		
def unregisterface(mesh, conn, fi):
	f = mesh.faces[fi]
	for e in ((f[0],f[1]), (f[1],f[2]), (f[2],f[0])):
		if e in conn:	del conn[e]

def segmentsdict(line):
	segments = {}
	for i in range(len(line)-1):
		segments[(line[i],line[i+1])] = i
	return segments

def intersection_plane_plane(p0, p1, neigh=None):
	''' Return the intersection axis between two planes '''
	if not neigh:	neigh = p0[0]
	d = cross(p0[1], p1[1])
	if length(d) <= NUMPREC:	return None
	return vec3(
		inverse(transpose(dmat3(p0[1], p1[1], d))) 
		* dvec3(dot(p0[0],p0[1]), dot(p1[0],p1[1]), dot(neigh,d))
		), normalize(d)

def intersection_edge_plane(edge, axis, prec):
	''' Return the intersection point of an edge with a plane, or None if it doesn't exist 
		In case of an intersection at an extremity of the edge, return that edge point
	'''
	a, b = edge
	da = dot(axis[0]-a, axis[1])
	db = dot(axis[0]-b, axis[1])
	if abs(da) <= prec:		return a
	if abs(db) <= prec:		return b
	if da * db > 0:		return None		# the segment doesn't reach the plane
	edgedir = b-a
	proj = dot(edgedir, axis[1])
	if abs(proj) <= prec:	return None	# the segment is parallel to the plane
	return a + da * edgedir / proj

def intersection_axis_face(axis, face, prec):
	''' Intersection between an axis and a triangle
		In case of intersection with an edge of the triangle, return None
	'''
	n = cross(face[1]-face[0], face[2]-face[0])
	unp = dot(n, axis[1])
	if abs(unp) <= prec:	return None
	p = axis[0] + dot(face[0]-axis[0], n) / unp * axis[1]
	for i in range(3):
		if dot(p - face[i], normalize(cross(n,face[i]-face[i-1]))) <= prec:
			return None
	return p



def removefaces(mesh, crit):
	''' Remove faces whose indices are present in faces, (for huge amount, prefer pass faces as a set) '''
	newfaces = typedlist(dtype=uvec3)
	newtracks = typedlist(dtype='I')
	for i in range(len(mesh.faces)):
		if not crit(i):
			newfaces.append(mesh.faces[i])
			newtracks.append(mesh.tracks[i])
	mesh.faces = newfaces
	mesh.tracks = newtracks
	
def removeedges(mesh, crit):
	''' Remove faces whose indices are present in faces, (for huge amount, prefer pass faces as a set) '''
	newedges = typedlist(dtype=uvec2)
	newtracks = typedlist(dtype='I')
	for i in range(len(mesh.edges)):
		if not crit(i):
			newedges.append(mesh.edges[i])
			newtracks.append(mesh.tracks[i])
	mesh.edges = newedges
	mesh.tracks = newtracks

#def facesurf(mesh, fi):
	#o,x,y = mesh.facepoints(fi)
	#return length(cross(x-o, y-o))

#def faceangle(mesh, fi):
	#o,x,y = mesh.facepoints(fi)
	#return length(cross(normalize(x-o), normalize(y-o)))

def faceheight(mesh, fi):
	f = mesh.facepoints(fi)
	m = inf
	for i in range(3):
		a,b = f[i]-f[i-1], f[i]-f[i-2]
		if not dot(b,b):	return 0
		m = min(m, dot(a,a) - dot(a,b)**2/dot(b,b))
	return sqrt(max(0,m))
	
def faceheight(mesh, fi):
	f = mesh.facepoints(fi)
	m = inf
	for i in range(3):
		l = length(f[i-2]-f[i])
		if not l:	return 0
		h = length(cross(f[i-2]-f[i], f[i-1]-f[i])) / l
		if h < m:	m = h
	return m



@chamfer.register(Mesh)
def mesh_chamfer(mesh, edges, **cutter):
	''' 
		Create a chamfer on the given suite of points, create faces are planes.
		cutter is described in function planeoffsets()
	'''
	# cut faces
	segments = mesh_edgecut(mesh, edges, **cutter)
	
	# create junctions
	group = len(mesh.groups)
	mesh.groups.append(None)
	pts = mesh.points
	# edges for corners surfaces
	corners = {}
	def cornerreg(c, edge):
		if c not in corners:	corners[c] = []
		corners[c].append(edge)
	
	for e,s in segments.items():
		if not s:	continue
		# corner cuts
		if isinstance(e, Number):
			if e not in corners:	corners[e] = []
			corners[e].extend(s)
		# edge cuts
		else:
			# assemble the intersection segments in a single outline
			frags = suites(s)
			if len(frags) == 1:		
				lp = frags[0]
				cornerreg(e[0], (lp[0],lp[-1]))
			elif len(frags) == 2:	
				lp = frags[0] + frags[1]
				# get the sides in the edge direction
				l,r = frags
				if dot(pts[l[-1]]-pts[l[0]], pts[e[1]]-pts[e[0]]) < 0 or dot(pts[r[-1]]-pts[r[0]], pts[e[1]]-pts[e[0]]) > 0:
					l,r = r,l
				cornerreg(e[0], (l[0], r[-1]))
				cornerreg(e[1], (r[0], l[-1]))
			else:
				raise MeshError('a cutted edge has more than 2 cutted sides')
			# triangulate cutted edge
			faces = triangulation_outline(Wire(mesh.points, lp)).faces
			mesh.faces.extend(faces)
			mesh.tracks.extend([group]*len(faces))
	
	for c,s in corners.items():
		# triangulate cutted corner
		faces = triangulation_outline(Wire(mesh.points, suites(s)[0])).faces
		mesh.faces.extend(faces)
		mesh.tracks.extend([group]*len(faces))

@filet.register(Mesh)
def mesh_filet(mesh, edges, resolution=None, **cutter):
	''' Create a chamfer on the given suite of points, create faces are planes.
		cutter is described in function planeoffsets()
	'''
	if isinstance(edges, Web):	edges = edges.edges
	
	edges = [edgekey(*e)	for e in edges]
	conn = connef(mesh.faces)
	enormals = {}
	for e in edges:
		enormals[e] = (
			mesh.facenormal(conn[e]),
			mesh.facenormal(conn[(e[1],e[0])]),
			)
	# cut faces
	segments = mesh_edgecut(mesh, edges, conn, **cutter)
	mesh.check()
	conn = connef(mesh.faces)
	pts = mesh.points
	
	# edges for corners surfaces
	corners = {}
	def cornerreg(c, edge):
		if c not in corners:	corners[c] = []
		corners[c].append(edge)
	
	# create junctions
	normals = {}	# face normal at each point
	junctions = []
	ends = []
	for e,s in segments.items():
		if not s:	continue
		# corner cuts
		if isinstance(e, Number):	
			if e not in corners:	corners[e] = []
			corners[e].extend(s)
		# edge cuts
		else:
			# assemble the intersection segments in a single outline
			frags = suites(s)
			
			# one loop:  filet extremity
			if len(frags) == 1:
				lp = frags[0]
				# find a separation to put a junction between 2 sides
				side = noproject(pts[lp[0]]-pts[lp[-1]], normalize(pts[e[1]] - pts[e[0]]))
				for i in range(1,len(lp)):
					if dot(pts[lp[i-1]]-pts[e[0]], side) * dot(pts[lp[i]]-pts[e[0]], side) < 0:
						l,r = lp[:i], lp[i:]
						break
				cornerreg(e[0], (lp[0],lp[-1]))
				if (lp[i], lp[i-1]) in conn:
					ends.append((lp[i-1], lp[i]))
			# two parts: cutted edge
			elif len(frags) == 2:
				l,r = frags
				if dot(pts[l[-1]]-pts[l[0]], pts[e[1]]-pts[e[0]]) < 0 or dot(pts[r[-1]]-pts[r[0]], pts[e[1]]-pts[e[0]]) > 0:
					l,r = r,l
				cornerreg(e[0], (l[0], r[-1]))
				cornerreg(e[1], (r[0], l[-1]))
			else:
				raise MeshError('a cutted edge has more than 2 cutted sides')
			
			# match the curves
			r.reverse()
			match = list(match_length(Wire(pts,l), Wire(pts,r)))
			# prepare tangents
			ll = Wire(pts,l).length()
			x = 0.
			for i in range(len(match)):
				if i:	x += distance(pts[match[i-1][0]], pts[match[i][0]]) / ll
				# get the tangents from the triangle between the cuts and the cutting edge
				l,r = match[i]
				o = interpol1(pts[e[1]], pts[e[0]], x)
				plane = cross(pts[r]-o, pts[l]-o)
				normals[l] = enormals[e][1] + normals.get(l, 0)
				normals[r] = enormals[e][0] + normals.get(r, 0)
			junctions.append(match)
		
	# normals neighbooring corners
	for e,s in corners.items():
		for a,b in s:
			if (b,a) in conn:
				n = mesh.facenormal(conn[(b,a)])
				normals[b] = normals.get(b, 0) + n
				normals[a] = normals.get(a, 0) + n
	# normals neighbooring cut ends
	for a,b in ends:
		n = mesh.facenormal(conn[(b,a)])
		normals[a] = noproject(normals[a], n)
		normals[b] = noproject(normals[b], n)
	# normalize all contributions
	for k,n in normals.items():
		normals[k] = normalize(n)
	
	# compute resolution based on adjacent faces to edges
	div = 0
	c = 0
	for match in junctions:
		for a,b in match:
			div += settings.curve_resolution(
						arclength(a,b,normals[a],normals[b]), 
						anglebt(normals[a], normals[b]),
						resolution)
			c += 1
	div //= c
	
	new = Mesh()
	# round cutted corner	
	for e,s in corners.items():
		if len(s) < 3:	continue
		# assemble the intersection segments in a single outline
		lp = suites(s)[0]
		if lp[-1] == lp[0]:	lp.pop()	# remove the redundant point if there is (not all the time due to some BUG)
		new += tangentcorner(pts, lp, normals, div)
	# fill gap between existing surface and round extremities
	for edge in ends:
		n = mesh.facenormal(conn[(edge[1],edge[0])])
		new += tangentend(pts, edge, normals, div)
	# round cutted edges
	for match in junctions:
		# put a junction
		new += tangentjunction(pts, match, normals, div)
	
	new.groups = [None]
	new.tracks = typedlist.full(0, len(new.faces), 'I')
	mesh += new
	mesh.mergeclose()

	
	
	
	
# ---- web operations -----
from .nprint import nprint
	
def web_cut(web, start, cutplane, conn, prec, removal):
	''' Propagation cut for a point 
		
		:start:		the edge or point to start propagation from
		:cutplane:	the plane cutting the faces. Its normal must be oriented toward the propagation area.
	'''
	intersections = []
	
	# propagate from point to point
	seen = set()
	front = list(conn[start])
	while front:
		ei = front.pop()
		if ei in seen:	continue
		seen.add(ei)
		e = web.edges[ei]
		
		# look intersection with the cutplane
		ep = web.edgepoints(e)
		side = [dot(p-cutplane[0], cutplane[1])		for p in ep]
		if side[0]*side[1] <= prec:
			p = intersection_edge_plane(ep, cutplane, prec)
			
			pi = len(web.points)
			web.points.append(p)
			intersections.append(pi)
			if side[0] <= prec:
				if distance2(p,ep[0]) < prec**2:
					pi = e[0]
				web.edges[ei] = (pi, e[1]) 
				web.edges.append((e[0], pi))
				web.tracks.append(web.tracks[ei])
				conn.discard(e[0], ei)
				conn.add(e[0], len(web.edges)-1)
			else:
				if distance2(p,ep[1]) < prec**2:
					pi = e[1]
				web.edges[ei] = (e[0], pi)
				web.edges.append((pi, e[1]))
				web.tracks.append(web.tracks[ei])
				conn.discard(e[1], ei)
				conn.add(e[1], len(web.edges)-1)
		else:
			# propagate
			for p in e:
				front.extend(conn[p])
		
	removal.update(seen)
	return intersections

@edgecut.register(Web)
def web_edgecut(web, points, conn=None, prec=None, removal=None, **cutter):
	if conn is None:	conn = connpe(web.edges)
	if prec is None:	prec = web.precision()
	if isinstance(points, Wire):	points = points.indices
	
	removal = set()
	cutter = interpretcutter(**cutter)
	pts = web.points
	intersections = {}
	
	for pi in points:
		dirs = []
		for ni in conn[pi]:
			e = web.edges[ni]
			dirs.append(normalize( pts[pi] - pts[e[e[0] == pi]] ))
		if not dirs:  # the cut has removed everything connected
			continue
		normal = normalize(sum(dirs))
		c = sum(dot(normal,d) for d in dirs) / len(dirs)
		s = sqrt(max(0, 1-c**2))
		x,_,_ = dirbase(normal)
		offset = cutter(c*x + s*normal, -c*x + s*normal)
		if dot(offset, normal) > 0:	offset = -offset
		plane = (pts[pi]+offset, -normalize(offset))
		
		intersections[pi] = web_cut(web, pi, plane, conn, prec, removal)
	
	removeedges(web, lambda i: i in removal)	
	return intersections
	
@chamfer.register(Web)
def web_chamfer(web, points, **cutter):
	holes = web_edgecut(web, points, **cutter)

	g = len(web.groups)
	web.groups.append(None)
	
	conn = connpe(web.edges)
	def link(e):
		if web.edges[next(conn[e[0]])][0] == e[0]:	return (e[1],e[0])
		else:                                  		return e
	
	for cuts in holes.values():
		if len(cuts) == 2:
			web.edges.append(link(tuple(cuts)))
			web.tracks.append(g)
		else:
			pi = len(web.points)
			web.points.append(sum(web.points[c] for c in cuts) / len(cuts))
			web.edges.extend( link((c,pi))  for c in cuts )
			web.tracks.extend( [g] * len(cuts) )
	
@filet.register(Web)
def web_filet(obj, points, resolution=None, **cutter):
	holes = web_edgecut(obj, points, **cutter)

	conn = connpe(obj.edges)
	pts = obj.points
	div = 6
	
	g = len(obj.groups)
	obj.groups.append(None)
	
	def tangentat(p):
		''' Find the tangent axis to the given cut point, 
			the third element is whether the original edge direction is opposite to the given tangent 
		'''
		e = obj.edges[next(conn[p])]
		if e[0] == p:	return pts[p], normalize(pts[p] - pts[e[1]]), True
		else:			return pts[p], normalize(pts[p] - pts[e[0]]), False
	
	for cuts in holes.values():
		if len(cuts) == 2:
			# place an arc between the two ends
			s0 = tangentat(cuts[0])
			s1 = tangentat(cuts[1])
			corner = web(tangentarc(s0[:2], s1[:2], resolution))
			if s0[2]:	corner = corner.flip()
			obj += corner
			
		else:
			# find an approximate sphere center
			tangents = [tangentat(i)  for i in cuts]
			normal = normalize(sum(t[1]  for t in tangents))
			med = sum(pts[i]    for i in cuts) / len(cuts)
			center = sum(pts[i] + unproject(med-pts[i], normal)   for i in cuts)
			radius = sum(distance(pts[i], center)   for i in cuts)
			
			# place summit
			pi = len(pts)
			top = center + normal*radius
			pts.append(top)
			# place arcs
			for t in tangents:
				u = (top, noproject(t[0]-top, normal))
				corner = web(tangentarc(t[:2], u, resolution))
				if t[2]:	corner = corner.flip()
				obj += corner


				
# ---- wire operations -----

@edgecut.register(Wire)
def wire_edgecut(wire: Wire, points, **cutter):
	if isinstance(points, Wire):
		points = points.indices
	if not isinstance(points, set):
		points = set(points)
	prec = wire.precision()

	cutter = interpretcutter(**cutter)
	if not wire.tracks:
		wire.tracks = typedlist.full(0, len(wire.indices), "I")
	g = len(wire.groups)
	wire.groups.append(None)

	cuts = []
	closed = wire.indices[0] == wire.indices[-1]
	
	for index, origin in enumerate(wire.indices):
		if origin not in points:
			continue
		# get point location in the wire
		if not closed:
			if origin == wire.indices[0] or origin == wire.indices[-1]:
				raise MeshError("a chamfer/filet cannot have only one side")

		# compute cut plane
		ref = wire[index-1] - wire[index]
		p0, p1 = wire[index:index+2]
		# handle special case of closed wire
		prior_p = wire[index-1]
		if prior_p == p0:
			prior_p = wire[index-2]
		
		t0, t1 = normalize(p0 - prior_p), normalize(p1 - p0)
		axis = cross(t0, t1)
		offset = cutter(normalize(cross(axis, t0)), normalize(cross(axis, t1)))
		if dot(offset, t0) > 0:
			offset = -offset
		cutplane = (p0 + offset, -normalize(offset))

		def find_intersection(it):
			for ii0, ii1 in it:
				if ii0 == -1: # handle closed wires
					ii0 = -2					
				p0, p1 = wire[ii0], wire[ii1]
				ps = intersection_edge_plane((p0, p1), cutplane, prec)
				if ps:
					if isfinite(ps):
						return ii1, ps
			raise ValueError("no intersection found")			

		# propagate backward
		back_range = range(index, -closed, -1)
		back_it = zip(map(lambda x: x-1, back_range), back_range)
		iiback, ps = find_intersection(back_it)
		cuts.append((iiback, iiback + 1))  # point interval including start and excluding end

		# propagate forward
		wlen = len(wire.indices)
		forward_range = range(index, wlen-1)
		forward_it = zip(forward_range, map(lambda x: x+1, forward_range))
		iiforward, pe = find_intersection(forward_it)

		# update wire
		m = len(wire.points)
		wire.points.append(ps)
		wire.points.append(pe)

		wire.indices[iiback: iiforward] = [m, m + 1]
		wire.tracks[iiback: iiforward] = [g, wire.tracks[iiforward-1]]

		if origin == wire.indices[-1]:
			wire.indices[-1] = wire.indices[0]
			wire.tracks[-1] = wire.tracks[0]

	return cuts
		
@chamfer.register(Wire)
def wire_chamfer(wire, points, **cutter):
	wire_edgecut(wire, points, **cutter)

@filet.register(Wire)
def wire_filet(wire, points, resolution=None, **cutter):
	closed = wire.indices[0] == wire.indices[-1]

	cuts = wire_edgecut(wire, points, **cutter)
	g = len(wire.groups) - 1
	wire.groups[g] = None

	cuts.reverse()
	for ii0, ii1 in cuts:
		p0 = wire[ii0]
		p1 = wire[ii1]
		tpii0 = ii0 - 1  # indices index if prior point
		tpii1 = ii1 + 1  # indices index if prior point
		if tpii0 < 0:
			tpii0 = -2

		t0 = normalize(p0 - wire[tpii0])
		t1 = normalize(p1 - wire[tpii1])
		old_l = len(wire.points)
		wire.points.extend( tangentarc((p0,t0), (p1,t1), resolution))
		wire.indices[ii0:ii0+2] = range(old_l, len(wire.points))
		wire.tracks[ii0:ii0+1] = [g] * (len(wire.points)-old_l-1)
		if closed and ii0==0:
			wire.indices[-1] == wire.indices[0]


		
		
# --- mesh tangent generation functions ----

def tangentarc(s0, s1, resolution=None):
	''' Create a tangent curve to both given axis. 
		return the curve as a list of points 
	'''
	p0, t0 = s0
	p1, t1 = s1
	z = cross(t0,t1)
	factor = arclength(p0, p1, cross(z,t0), cross(t1,z))
	s0 = (p0, factor*t0)
	s1 = (p1, factor*t1)
	div = settings.curve_resolution(factor, anglebt(t0,-t1), resolution)
	return [ interpol2(s0, s1, j/(div+1))   for j in range(div+2) ]

def tangentend(points, edge, normals, div):
	''' Join a tangent surface resulting of `tangentcorner` or `tangentjunction` to a straight edge e 
		`normals` is the same dict as for tangentcorner and tangentjunction
	'''
	l, r = edge
	pl, pr = points[l], points[r]
	dist = arclength(points[l], points[r], normals[l], normals[r])
	e = points[r]-points[l]
	tl = normalize(noproject( e, normals[l])) * dist
	tr = normalize(noproject(-e, normals[r])) * dist
	
	def infos():
		for i in range(div+2):
			x = i/(div+1)
			yield interpol2((pl,tl), (pr,tr), x),  interpol1(pl,pr, x)
	return blenditer(infos(), 0, interpol1)
	

def tangentcorner(pts, lp, normals, div):
	''' Create a rounded surface tangent to the loop given
		`normals` is a dict {point: normal}
	'''
	new = Mesh()
	# compute the common points to all triangular regions
	n = normalize(sum(normals[i] for i in lp))
	med = sum(pts[i] for i in lp) / len(lp)
	center = sum(pts[i] + unproject(med-pts[i], normals[i])	for i in lp) / len(lp)
	radius = sum(distance(pts[i], center) for i in lp) / len(lp)
	# place a central point
	p = len(pts)
	pts.append(center + n*radius)
	normals[p] = n
	# triangulate
	for i in range(len(lp)):
		a,b,c = lp[i-1], lp[i], p
		new += gt.icosurface(
					(pts[a],pts[b],pts[c]), 
					(normals[a],normals[b],normals[c]),
					resolution=('div',div))
	return new

def tangentjunction(points, match, normals, div):
	''' Create a surface joining the given couples of points, tangent to the two sides
		`normals` is a dict {point: normal}
	'''
	def infos():
		# estimate normals
		for l,r in match:
			dist = arclength(points[l], points[r], normals[l], normals[r])
			e = points[r]-points[l]
			tl = normalize(noproject( e, normals[l])) * dist
			tr = normalize(noproject(-e, normals[r])) * dist
			yield (points[r], tr), (points[l], tl)
	
	return blenditer(infos(), div, interpol2)
